/*
 *	springboard.h
 *
 *	A spring board allows to dynamically resize views using constraints
 *	which are very similar to physical springs (these may be non-linear).
 *
 *	(C) Copyright 1994-1997, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef	_OPaC_SPRINGBOARD_H_
#define	_OPaC_SPRINGBOARD_H_

#ifndef _OPaC_RECT_H_
#include <OPaC/rect.h>
#endif

OP_INTERFACE (springboard.h);

/****************************************************************/

/*
 *	The Spring Band maintains a vertical or horizontal collection
 *	of views connected by springs.
 */

class OP_PUBLIC OPSpringBand : public OPObject
{

protected:
	Bool				is_vertical;			//	TRUE => vertical band
	Bool				is_meta_ready;			//	TRUE => the meta spring is ready
	
	OPSpringRef			meta_spring;			//	equivalent to the individual springs
	OPSpringArray		spring;					//	there are 2*n+1 springs
	OPViewArray			view;					//	there are n views
	
protected:
	OPSpringBand ();
	virtual ~ OPSpringBand ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */

	virtual void BuildMetaSpring ();
	virtual void RemoveAll ();
	
public:
	virtual void Delete ();								/* OVERRIDDEN */
	
	virtual void SetVertical (Bool yes = TRUE);
	virtual Bool IsVertical () const;
	
	virtual void SpringChanged ();
	virtual void Resize (OPRectIn rect);
	
	virtual Count CountViews () const;
	virtual Count CountSprings () const;

	virtual Bool InsertView (Count index, OPViewIn view);
	virtual Bool RemoveView (Count index);

	virtual Bool GetNthView (Count index, OPViewOut view) const;
	virtual Bool GetNthSpring (Count index, OPSpringOut spring) const;
	
	virtual Bool FindView (OPViewIn view, CountOut index) const;
	
	virtual Bool FindSpringBeforeView (Count index, OPSpringOut spring) const;
	virtual Bool FindSpringOfView (Count index, OPSpringOut spring) const;
	virtual Bool FindSpringAfterView (Count index, OPSpringOut spring) const;
	
	virtual Bool FindSpringBeforeView (OPViewIn view, OPSpringOut spring) const;
	virtual Bool FindSpringOfView (OPViewIn view, OPSpringOut spring) const;
	virtual Bool FindSpringAfterView (OPViewIn view, OPSpringOut spring) const;
	
	OP_DEFINE_CLASS (OPSpringBand, OPObject);
};

/****************************************************************/

/*
 *	The spring board is just a collection of spring bands. The
 *	views will be moved and resized sequentially, from band to
 *	band.
 */

class OP_PUBLIC OPSpringBoard : public OPObject
{
	
protected:
	OPSpringBandArray	band;					//	array of bands

protected:
	OPSpringBoard ();
	virtual ~ OPSpringBoard ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RemoveAll ();
	
public:
	virtual void Delete ();								/* OVERRIDDEN */
	
	virtual void Resize (OPRectIn rect);
	
	virtual Count CountBands () const;
	
	virtual Bool InsertBand (Count index, OPSpringBandIn band);
	virtual Bool RemoveBand (Count index);
	
	virtual Bool GetNthBand (Count index, OPSpringBandOut band) const;
	virtual Bool FindBand (OPSpringBandIn band, CountOut index) const;
	
	OP_DEFINE_CLASS (OPSpringBoard, OPObject);
};

/****************************************************************/

#endif /* _OPaC_SPRINGBOARD_H_ */
