/*
 *	window.h
 *
 *	Base Window class definition. The window class contains every other widget
 *	and provides an interface to the underlying window system.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef _OPaC_WINDOW_H_
#define	_OPaC_WINDOW_H_

#ifndef _OPaC_OPEXPORT_H_
#include <OPaC/opexport.h>
#endif

OP_INTERFACE (window.h);

#ifndef _OPaC_VIEW_H_
#include <OPaC/view.h>
#endif

#ifndef _OPaC_WINSTYLE_H_
#include <OPaC/winstyle.h>
#endif

#ifndef _OPaC_STRING_H_
#include <OPaC/string.h>
#endif

#ifndef _OPaC_XTWIDGET_H_
#include <OPaC/xtwidget.h>
#endif

/****************************************************************/

/*
 *	OPaC Window class provides a portable description of
 *	a window. It is basically a view with some additional
 *	methods to handle window related events.
 */

class OP_PUBLIC OPWindow : public OPView, public OPxTitleWidget
{
	
	static OPAtomRef	atom_title;				//	atom "Title" used for queries
	
protected:
	OPAppRef			app;					//	the application
	OPDataSlotRef		slot;					//	the associated data slot (if any)
	OPGrailWindowRef	grail;					//	Grail window description
	
	Bool				is_open;				//	TRUE => window may be drawn into
	Bool				is_visible;				//	TRUE => window is visible on screen
	Bool				is_reduced;				//	TRUE => window has been reduced
	
	OPWindowStyle		style;					//	window style
	OPStringRef			title;					//	window title (name)
	OPViewRef			reduced;				//	reduced view
	
	OPViewRef			default_active;			//	default active view
	OPViewRef			current_active;			//	current active view
	Bool				current_active_ok;		//	TRUE => current active view is enabled
	
	OPHashTableRef		view_collection;		//	table referring to the existing views
	OPHashTableRef		view_overlays;			//	table referring to overlay views
	OPHashTableRef		key_alt_collection;		//	table referring to shortcuts
	
	OPViewRef			cursor_view;			//	current view containing graph.cursor
	OPViewRef			drag_initial_view;		//	view where the dragging began
	
protected:
	OPWindow ();
	virtual ~ OPWindow ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
	virtual void ChangeGeometry (OPRectIn);				/* OVERRIDDEN */
	virtual void AwakeMyself ();
	
	virtual OPView* ReturnObjectAsView ();
	virtual void DrawOverlays (OPGrailPortIn port, OPRectIn rect);
	
public:
	virtual void Delete ();								/* OVERRIDDEN */
	
	virtual void Awake ();								/* OVERRIDDEN */
	virtual void AwakeFromApp (OPAppIn app);			/* OVERRIDDEN */
	virtual void AwakeFromObject (OPObjectIn object);	/* OVERRIDDEN */
	
	//	The window's style has to be set before opening it.
	
	virtual Bool SetStyle (OPWindowStyle style);
	virtual OPWindowStyle ReturnStyle () const;
	
	//	A window's name may be changed any time :
	
	virtual Bool SetTitle (OPStringIn name);
	virtual Bool GetTitle (OPStringOut name) const;
	
	//	When the view gets reduced, it will have to display a small icon
	//	representing it :
	
	virtual Bool SetReducedView (OPViewIn view);
	virtual Bool GetReducedView (OPViewOut view) const;
	
	//	Opening a window will create the associated buffer and enable the
	//	painting operations, but the window still won't be visible on screen.
	
	virtual void Open ();
	virtual void Close ();
	virtual void Kill ();
	
	//	Showing a window will make it visible on screen (at normal size or in
	//	the reduced size). Hiding a window will remove it from screen, but not
	//	free the buffer associated to it; it will be redrawn very quickly if
	//	needed.
	
	virtual void ShowNormal ();
	virtual void ShowReduced ();
	virtual void Hide ();
	virtual void ActivateAsFrontWindow (Bool front = TRUE);
	
	virtual void SysChangedActive (Bool);
	virtual void SysChangedOpen (Bool);
	virtual void SysChangedVisible (Bool);
	virtual void SysChangedReduced (Bool);
	
	virtual Bool IsVisible () const;
	virtual Bool IsReduced () const;
	virtual Bool IsActive () const;
	
	//	When the view hierarchy changes, the window gets notification
	//	messages :
	
	virtual void NotifyInsertView (OPViewIn view);
	virtual void NotifyRemoveView (OPViewIn view);
	virtual void NotifyReparentView (OPViewIn view);
	
	virtual void NoLongerActive (OPViewIn view);

	virtual void NotifyInsertKeyAlt (OPControlIn ctrl);
	virtual void NotifyRemoveKeyAlt (OPControlIn ctrl);
	
	virtual void NotifyInsertOverlay (OPViewIn view);
	virtual void NotifyRemoveOverlay (OPViewIn view);
	
	//	Miscellanous :
	
	virtual Bool GetActiveView (OPViewOut view) const;
	virtual Bool GetCursorView (OPViewOut view) const;
	
	virtual void Redraw (OPRectIn box);
	virtual void UserChangedGeometry (OPRectIn box);
	
	virtual void GetFrameOffset (OPCoordOut dx, OPCoordOut dy);
	virtual void GetFrameOverhead (OPCoordOut dx, OPCoordOut dy);
	
	virtual const OPGrailPortRef& ReturnGrailPort () const;		/* OVERRIDDEN */
	virtual Bool Detect (OPCoord x, OPCoord y, OPViewOut view);	/* OVERRIDDEN */
	virtual void Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp = OP_DRAW_ALL);/* OVERRIDDEN */
	
	virtual void DispatchEvent (OPEventIn event);				/* OVERRIDDEN */
	
	virtual Bool SetMainResponder (OPResponderIn, Bool);
	virtual Bool GetMainResponder (OPResponderOut) const;
	
	virtual void RemoveMainResponder ();
	virtual void EnableMainResponder ();
	virtual void DisableMainResponder ();
	
	virtual Bool IsMainResponderOk () const;
	virtual const OPAppRef& ReturnApp () const;
	
	virtual Bool SetSlot (OPDataSlotIn slot);
	virtual Bool GetSlot (OPDataSlotOut slot);
	virtual const OPDataSlotRef& ReturnSlot ();
	
	virtual Bool SetDraggingView (OPViewIn view);
	virtual Bool GetDraggingView (OPViewOut view) const;
	
	//	Default event handlers :

	virtual void KeyDown (OPEventIn event);						/* OVERRIDDEN */
	virtual void KeyChange (OPEventIn event);					/* OVERRIDDEN */
	virtual void KeyUp (OPEventIn event);						/* OVERRIDDEN */
	virtual void ButtonPress (OPEventIn ev);					/* OVERRIDDEN */
	virtual void ButtonRelease (OPEventIn ev);					/* OVERRIDDEN */
	virtual void MotionNotify (OPEventIn event);				/* OVERRIDDEN */
	virtual void DragNotify (OPEventIn event);					/* OVERRIDDEN */
	virtual void Miscellanous (OPEventIn event);				/* OVERRIDDEN */
	virtual void EnterNotify (OPEventIn event);					/* OVERRIDDEN */
	virtual void LeaveNotify (OPEventIn event);					/* OVERRIDDEN */
	
	OP_DEFINE_CLASS (OPWindow, OPView);
};

#endif
