/*
 *	sd_grailpxm.cxx
 *
 *	GRAphical Interface Layer pixel map interface class. This is
 *	heavily used for double buffering. Implementation fot X11.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

// Modified by Laurent Bovet 01.02.98

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include "sd_interface.h"

#include <OPaC/grail.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENTATION (grail.h);
OP_IMPLEMENT_CLASS (OPGrailPixmap);

/****************************************************************/

/*
 *	Initialize the Grail Pixmap instance. This won't allocate
 *	any bitmap; do this using "Initialise" !
 */
	
OPGrailPixmap::OPGrailPixmap ()
{
	OP_METHOD (OPGrailPixmap, 0);
	
	this->pixmap = 0;
	this->dx     = 0;
	this->dy     = 0;
	this->depth  = 0;
	
	this->pixmap = OP_ALLOC_FROM_ZONE (this->ReturnMemZone (), OPaC_GrPixmap);
	
	this->pixmap->xpixmap = 0;
}



/*
 *	Destroy cleanly a Pixmap : this will free an existing
 *	bitmap and associated data.
 */

OPGrailPixmap::~OPGrailPixmap ()
{
	OP_METHOD (~OPGrailPixmap, 0);
	
	if (this->pixmap) {
		
		XFreePixmap(xdpy, this->pixmap->xpixmap);
		
		OP_FREE_FROM_ZONE (this->ReturnMemZone (), this->pixmap);
	}
}


/*
 *	Start drawing into the window. This will first draw into
 *	the buffering pixmap.
 */

void
OPGrailPixmap::BeginDraw (OPRectIn box, OPRelCoord margin)
{
	OP_METHOD (BeginDraw, 2);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPRelCoord, margin);
	OP_ASSERT (this->pixmap->xpixmap);
	super::BeginDraw (box, margin);
}



/*
 *	Start drawing into the pixel map. Tell the Grail Port
 *	which display context to use for drawing.
 */

void
OPGrailPixmap::BeginDraw (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2)
{
	OP_METHOD (BeginDraw, 4);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);
	OP_ASSERT (this->pixmap->xpixmap);
	super::BeginDraw (x1, y1, x2, y2);
}



/*
 *	End of drawing. Just propagate the message and remove
 *	the display context from the Grail Port.
 */

void
OPGrailPixmap::EndDraw ()
{
	OP_METHOD (EndDraw, 0);
	super::EndDraw ();
}



/*
 *	Initialise a new pixel map. If one already existed, just
 *	delete it and start anew. If "window" is not specified,
 *	the default screen will be used as model.
 */

void
OPGrailPixmap::Initialise (OPGrailPixmapIn window, OPCoord dx, OPCoord dy)
{
	OP_METHOD (Initialise, 3);
	OP_ARG (1, OPGrailPixmapIn, window);
	OP_ARG (2, OPCoord, dx);
	OP_ARG (3, OPCoord, dy);
	OP_ASSERT ((dx > 0) && (dy > 0));
	OP_ASSERT (this->is_drawing == FALSE);
	
	Window x_window = 0;
		
	if (window.IsValid ()) {
		x_window = window->ReturnPrivatePortInfo ()->xdrawable;
		XCopyGC(xdpy, window->ReturnPrivatePortInfo ()->xgc, 
						0xFFFF, // Copy all
						this->info->xgc);
	}
	else
		x_window = RootWindow(xdpy, DefaultScreen(xdpy));
	
	this->pixmap->xpixmap = 
		XCreatePixmap(xdpy, 
									x_window,
									(int) dx,
									(int) dy,
									XDefaultDepth(xdpy, DefaultScreen(xdpy)));

	this->info->xdrawable = this->pixmap->xpixmap;
	this->is_ready  = TRUE;
	this->dx        = dx;
	this->dy        = dy;
	this->depth     = XDefaultDepth(xdpy, DefaultScreen(xdpy));

}



/*
 *	Initialise a new pixel map with the specified depth.
 *	If one already existed, just delete it and start anew.
 */

void
OPGrailPixmap::Initialise (Count depth, OPCoord dx, OPCoord dy)
{
	OP_METHOD (Initialise, 3);
	OP_ARG (1, Count, depth);
	OP_ARG (2, OPCoord, dx);
	OP_ARG (3, OPCoord, dy);
	
	OP_ASSERT ((dx > 0) && (dy > 0));
	OP_ASSERT (this->is_drawing == FALSE);
	
	this->pixmap->xpixmap = XCreatePixmap(xdpy, 
								 RootWindow(xdpy, DefaultScreen(xdpy)),
								 (int) dx,
								 (int) dy,
								 depth ? (unsigned int) depth :
								 XDefaultDepth(xdpy, DefaultScreen(xdpy)));

	this->info->xgc = XCreateGC(xdpy, RootWindow(xdpy, DefaultScreen(xdpy)),
															0,
															NULL);

	this->info->xdrawable = this->pixmap->xpixmap;
	this->is_ready  = TRUE;
	this->dx        = dx;
	this->dy        = dy;
	this->depth     = depth;

}


/*
 *	Return the RGB values of a few horizontal pixels.
 */

Bool
OPGrailPixmap::SystemGetRGBPixels (OPCoord x, OPCoord y, OPCoord dx, Card8* r, Card8* g, Card8* b)
{
	OP_METHOD (SetRGBPixels, 6);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, Card8*, r);
	OP_ARG (5, Card8*, g);
	OP_ARG (6, Card8*, b);

	Card32 num = (Card32)(dx);
	
	// Create a table of X colors

	XImage* ximage;
	XColor table[(int)dx];	

  // Get an image for the line
	
	ximage = XGetImage(xdpy, this->pixmap->xpixmap, 
										 (int)x, (int)y, 
										 (unsigned int)dx, 1,
										 AllPlanes,
										 ZPixmap);

  // Fill in the table

	for(int i=0; i<num; i++)
		table[i].pixel = XGetPixel(ximage, i, 0);

	XDestroyImage(ximage);
		
  // Ask for rgb values

	XQueryColors(xdpy, 
							 XDefaultColormap(xdpy, DefaultScreen(xdpy)),
							 table,
							 (int) dx);

  // Fill in the given component arrays

	for(int i=0; i<num; i++) {
		r[i] = (Card8)(table[i].red >> 8);
		g[i] = (Card8)(table[i].green >> 8);
		b[i] = (Card8)(table[i].blue >> 8);
	}

	return TRUE;
}
