/*
 *	sd_colman.cxx
 *
 *	OPaC GRAIL color manager for 32-bit MS-Windows. This is the
 *	machine dependent part.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include "sd_interface.h"

#include <OPaC/grail.h>
#include <OPaC/new.h>

/****************************************************************/

OP_MORE_IMPLEMENTATION (OPGrailPort);

/****************************************************************/

/*
 *	Some macros used to initialise the color palette. That's
 *	easier than writing all the code by hand (and spares a
 *	friend function).
 */

#define	SetupPalette(_i,_p,_m,_r,_g,_b)							\
	do {														\
		int _i_ = _i;											\
		this->pal_entry[_i_].pixel = _p;						\
		this->pal_entry[_i_].match_enable = _m;					\
		this->pal_entry[_i_].r = _r;							\
		this->pal_entry[_i_].g = _g;							\
		this->pal_entry[_i_].b = _b;							\
		this->pal_entry[_i_].gray = ((_r)==(_g))&&((_g)==(_b));	\
	} while (0)

#define	SetupPaletteRGB(_i,_m,_r,_g,_b)							\
	SetupPalette (_i, PALETTERGB(_r,_g,_b), _m, _r, _g, _b)


/****************************************************************/

/*
 *	Initialise a color manager. Determine if it will have to
 *	use an indexed or a true color model.
 */	

OPColorMan::OPColorMan ()
{
	//	First, put every field into a known state (zero, zero, zero...)
	
	this->info = OP_ALLOC_FROM_ZONE (OPaC_sys_zone, OPaC_ColorMan);
	this->info->hpal = 0;
	
	this->depth = 0;
	
	this->pal_size  = 0;						//	no palette entries
	this->pal_gray  = 0;
	this->pal_cube  = 0;
	this->pal_named = 0;
	this->pal_entry = 0;
	
	this->cube_pos  = 0;
	
	this->named_num   = 0;						//	no named entries
	this->named_pos   = 0;
	this->named_entry = 0;
	
	this->color_model = OP_COLOR_MODEL_NONE;
	this->gray_model  = OP_GRAY_MODEL_MONO;
	
	HDC hdc = GetDC (NULL);						//	don't forget to release later on !
	
	//	Configure the Color Manager depending on WINDOW's current settings :
	
	int np     = GetDeviceCaps (hdc, PLANES);
	int nbpp   = GetDeviceCaps (hdc, BITSPIXEL);
	int depth  = np * nbpp;
	int ncolor = 1 << depth;
	
	//	Check if the display has a palette or if it is considered to be true
	//	color (or just black & white).
	
	Bool is_idx = (GetDeviceCaps (hdc, RASTERCAPS) & RC_PALETTE) ? TRUE : FALSE;
	
	int  nres   = (is_idx) ? GetDeviceCaps (hdc, NUMRESERVED) : 0;
	int  navail = ncolor - nres;
	
	
	
	if ( (is_idx == FALSE)				//	true color display
	  || (depth <= 4)					//	to few colors
	  || (navail < (15 + 6)) ) {		//	to few available colors in palette
		
		//	This handles the three following cases :
		//	
		//	- True color display => no palette needed.
		//	- Palette to small => use WINDOW's standard colors.
		//	- Not enough colors available for redefinition => use WINDOW's
		//	  standard colors.
		
		this->color_model = (depth < 4) ? OP_COLOR_MODEL_NONE
										: (depth < 10) ? OP_COLOR_MODEL_POOR
													   : OP_COLOR_MODEL_GOOD;
		
		this->gray_model  = (depth < 2) ? OP_GRAY_MODEL_MONO
										: (depth < 10) ? OP_GRAY_MODEL_POOR
													   : OP_GRAY_MODEL_GOOD;
		
		this->depth = depth;
		
	} else {
		
		//	A palette is available (at least 20 free entries). Define a smooth
		//	gray ramp (15 gray levels). If there is room left, allocate a 5 x 5 x 5
		//	color cube.
		
		int index = 0;
		int nbug  = 6;							//	6 WINDOWS colors (bugous WINDOWS...)
		int ngray = navail + 2 - nbug;			//	NB: black & white always available
		if (ngray > 31)  ngray = 31;			//	maximum 31 gray shades
		int ncube = navail - nbug - ngray;		//	rest is available for extra colors
		if (ncube > 125) ncube = 125;			//	limit to 5 x 5 x 5 color cube
		if (ncube < 125) ncube = 0;				//	maybe none...
		int nrest = 0;							//	currently no extra colors
		
		this->depth     = depth;
		this->pal_size  = nbug + ngray + ncube + nrest;
		this->pal_entry = OP_ALLOC_ARRAY_FROM_ZONE (OPaC_sys_zone, PaletteEntry, this->pal_size);
		this->pal_gray  = ngray;
		this->pal_cube  = ncube;
		this->pal_named = nrest;
		
		this->cube_pos  = nbug + ngray;
		this->named_pos = nbug + ngray + ncube;
		
		this->color_model = (ncube) ? OP_COLOR_MODEL_MEDIUM : OP_COLOR_MODEL_POOR;
		this->gray_model  = (ngray < 15) ? OP_GRAY_MODEL_POOR : OP_GRAY_MODEL_MEDIUM;
		
		int i;
		
		//	These 6 extra colors are needed by some WINDOWS icons and
		//	would produce much more palette updates if not defined here !
		//	
		//	Why do these Microsoft developers never think a bit before
		//	implementing mechanisms ? :-(
		
		SetupPaletteRGB (index++, FALSE, 0xBF, 0x00, 0x00);
		SetupPaletteRGB (index++, FALSE, 0x00, 0xBF, 0x00);
		SetupPaletteRGB (index++, FALSE, 0xBF, 0xBF, 0x00);
		SetupPaletteRGB (index++, FALSE, 0x00, 0x00, 0xBF);
		SetupPaletteRGB (index++, FALSE, 0xBF, 0x00, 0xBF);
		SetupPaletteRGB (index++, FALSE, 0x00, 0xBF, 0xBF);
		
		//	Setup gray ramp.
		
		for (i = 0; i < ngray; i++) {
			int gray = (int)((i * 256) / (ngray - 1));
			
			if (gray > 255) gray = 255;
			
			Card8 rr = gray;
			Card8 gg = gray;
			Card8 bb = gray;
			
			SetupPaletteRGB (index++, TRUE, rr, gg, bb);
		}
		
		int r, g, b;
		
		for (r = 0; r < 5; r++) {
			for (g = 0; g < 5; g++) {
				for (b = 0; b < 5; b++) {
					
					Card8 rr = r * 64; if (r == 4) rr = 0xFF;
					Card8 gg = g * 64; if (g == 4) gg = 0xFF;
					Card8 bb = b * 64; if (b == 4) bb = 0xFF;
					
					SetupPaletteRGB (index++, TRUE, rr, gg, bb);
				}
			}
		}
		
		for (i = 0; i < nrest; i++)
			SetupPaletteRGB (index++, FALSE, 0x00, 0x00, 0x00);
		
		
		//	The palette has been initialised. Now we will have to tell WINDOWS
		//	about our needs.
		
		Count  bytes = sizeof (LOGPALETTE)
					   + (Count)(this->pal_size) * sizeof (PALETTEENTRY);
		Card8* mem   = OP_ALLOC_ARRAY_FROM_ZONE (OPaC_temp_zone, Card8, bytes);
		
		LOGPALETTE* logical = (LOGPALETTE*)(mem);
		
		logical->palNumEntries = (int)(this->pal_size);
		logical->palVersion    = 0x300;
		
		for (i = 0; i < (int)(this->pal_size); i++) {
			
			logical->palPalEntry[i].peRed   = this->pal_entry[i].r;
			logical->palPalEntry[i].peGreen = this->pal_entry[i].g;
			logical->palPalEntry[i].peBlue  = this->pal_entry[i].b;
			logical->palPalEntry[i].peFlags = PC_NOCOLLAPSE;
			
			if ( (this->pal_entry[i].r == 0x00)
			  && (this->pal_entry[i].g == 0x00)
			  && (this->pal_entry[i].b == 0x00) )
				logical->palPalEntry[i].peFlags = 0;	//	use standard black
			
			if ( (this->pal_entry[i].r == 0xFF)
			  && (this->pal_entry[i].g == 0xFF)
			  && (this->pal_entry[i].b == 0xFF) )
				logical->palPalEntry[i].peFlags = 0;	//	use standard white
		}
		
		this->info->hpal = CreatePalette (logical);
		
		OP_FREE_FROM_ZONE (OPaC_temp_zone, mem);
	}
	
	ReleaseDC (NULL, hdc);
	
	this->ComputeNearestColor ();
}



/* 
 *	Clean up the color manager.
 */

OPColorMan::~OPColorMan ()
{
	if (this->info && this->info->hpal) {
		DeleteObject (this->info->hpal);
	}
	
	if (this->info) {
		OP_FREE_FROM_ZONE (OPaC_sys_zone, this->info);
	}
	
	if (this->pal_entry) {
		OP_FREE_FROM_ZONE (OPaC_sys_zone, this->pal_entry);
	}
}



/*
 *	Return the pixel value for the RGB components. This
 *	will be used to get a true color pixel representation.
 */

OPColorPixel
OPColorMan::MakeTrueColor (Card8 r, Card8 g, Card8 b) const
{
	return (OPColorPixel)(RGB (r, g, b));
}



/*
 *	Return the color components from a pixel value.
 */

Bool
OPColorMan::FindColor (OPColorPixel p, Card8Out r, Card8Out g, Card8Out b) const
{
	OP_METHOD (FindColor, 4);
	OP_ARG (1, OPColorPixel, p);
	OP_ARG (2, Card8Out, r);
	OP_ARG (3, Card8Out, g);
	OP_ARG (4, Card8Out, b);
	
	Card32 pix32 = (Card32)(p);
	
	switch (pix32 & 0xFF000000) {
		case 0x01000000:
			{
				PALETTEENTRY entry = { 0, 0, 0, 0 };
				GetPaletteEntries (this->info->hpal, (Card8)(pix32), 1, &entry);
				r = entry.peRed;
				g = entry.peGreen;
				b = entry.peBlue;
			}
			break;
		
		default:
			r = (Card8)((pix32 & 0x0000FF) >> 0);
			g = (Card8)((pix32 & 0x00FF00) >> 8);
			b = (Card8)((pix32 & 0xFF0000) >> 16);
			break;
	}
	
	return TRUE;
}
