/*
 *	sd_grailcur.cxx
 *
 *	CAUTION: For 32-bit Windows only !
 *
 *	GRAphical Interface Layer for Cursor shape.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/


#include "sd_interface.h"

#include <OPaC/grail.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPGrailCursor);

/****************************************************************/
	
extern HCURSOR global_OPaC_cursor;

/****************************************************************/

/*
 *	Create an empty graphical cursor.
 */

OPGrailCursor::OPGrailCursor ()
{
	OP_METHOD (OPGrailCursor, 0);
	
	this->cursor = OP_ALLOC_FROM_ZONE (this->ReturnMemZone (), OPaC_GrCursor);
	this->cursor->hcursor = 0;
	
	this->foreground = 0;
	this->background = 0;
	
	this->x_foreground = 0;
	this->y_foreground = 0;
	
	this->x_background = 0;
	this->y_background = 0;
	
	this->dx = 0;
	this->dy = 0;
	
	this->hot_x = 0;
	this->hot_y = 0;
}



/*
 *	Free & destroy a graphical cursor.
 */

OPGrailCursor::~OPGrailCursor ()
{
	OP_METHOD (~OPGrailCursor, 0);
	
	if (this->cursor && this->cursor->hcursor) {
		
		HCURSOR current = GetCursor ();
		
		if (current == this->cursor->hcursor) {
			SetCursor (0);
		}
		
		DestroyCursor (this->cursor->hcursor);
	}
	
	OP_FREE_FROM_ZONE (this->ReturnMemZone (), this->cursor);
}



/*
 *	Save the cursor. This cannot be done, since this would
 *	lead to machine dependent archives.
 */

void
OPGrailCursor::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	OP_ABORT ("Trying to save Grail Cursor");
}



/*
 *	Restore the cursor. This cannot be done.
 */

void
OPGrailCursor::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	OP_ABORT ("Trying to restore Grail Cursor");
}



/*
 *	Initialise a cursor from the specified pixel maps. The
 *	background will be opaque wherever it is set to black.
 *	The foreground will be painted wherever a black back-
 *	ground has been drawn.
 *
 *	The cursor definition is the same as on the Smaky
 *	operating system and has been chosen because it is
 *	quite easy to draw the masks manually.
 */

void
OPGrailCursor::Initialise (OPCoord hotx, OPCoord hoty,
						   OPCoord dx, OPCoord dy,
						   OPGrailPixmapIn fg, OPCoord xf, OPCoord yf,
						   OPGrailPixmapIn bg, OPCoord xb, OPCoord yb)
{
	OP_METHOD (Initialise, 10);
	OP_ARG (1, OPCoord, hotx);
	OP_ARG (2, OPCoord, hoty);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, OPCoord, dy);
	OP_ARG (5, OPGrailPixmapIn, fg);
	OP_ARG (6, OPCoord, xf);
	OP_ARG (7, OPCoord, yf);
	OP_ARG (8, OPGrailPixmapIn, bg);
	OP_ARG (9, OPCoord, xb);
	OP_ARG (10, OPCoord, yb);
	
	if (this->cursor && this->cursor->hcursor) {
		
		HCURSOR current = GetCursor ();
		
		if (current == this->cursor->hcursor) {
			SetCursor (0);
		}
		
		DestroyCursor (this->cursor->hcursor);
	}
	
	//	When the cursor size is invalid, don't use any cursor at all. This
	//	will force the use of a default (system) arrow cursor.
	
	if ( (dx * dy) == 0) {
		
		this->cursor->hcursor = 0;
		
		this->foreground = 0;
		this->background = 0;
		
		this->x_foreground = 0;
		this->y_foreground = 0;
		
		this->x_background = 0;
		this->y_background = 0;
		
		this->dx = 0;
		this->dy = 0;
		
		this->hot_x = 0;
		this->hot_y = 0;
		
		return;
	}
	
	
	//	The grail pixmap reference and both memory chunks have to be recorded
	//	as an automatic variable so that these will be properly destroyed on
	//	exception.
	
	OPGrailPixmapRef mono;
	Card8* and_bits = 0;
	Card8* xor_bits = 0;
	
	int mx = GetSystemMetrics (SM_CXCURSOR);
	int my = GetSystemMetrics (SM_CYCURSOR);
	
	if (dx > mx) dx = mx;
	if (dy > my) dy = my;
	
	Card32 bytes = mx * my / 8;
	
	and_bits = OP_ALLOC_ARRAY_FROM_ZONE (OPaC_temp_zone, Card8, bytes);
	xor_bits = OP_ALLOC_ARRAY_FROM_ZONE (OPaC_temp_zone, Card8, bytes);
	mono     = OP_NEW_FROM_ZONE (OPaC_temp_zone, OPGrailPixmap);
	
	//	Convert both foreground and background to monochrome. Real bitmaps
	//	are needed for the cursor definition.
	
	mono->Initialise (1, mx, my);

	OPCoord mmx = mx-1;
	OPCoord mmy = my-1;
	
	mono->BeginDraw (0, 0, mmx, mmy);
	mono->SetForeColor ((OPColorPixel)(RGB(255,255,255)));
	mono->FillRect (0, 0, mmx, mmy);
	mono->CopyRect (fg, xf, yf, 0, 0, dx - 1, dy - 1);
	mono->EndDraw ();
	
	GetBitmapBits (mono->ReturnPrivatePixmapInfo ()->hbitmap, bytes, xor_bits);
	
	mono->BeginDraw (0, 0, mmx, mmy);
	mono->SetForeColor ((OPColorPixel)(RGB(255,255,255)));
	mono->FillRect (0, 0, mmx, mmy);
	mono->CopyRect (bg, xb, yb, 0, 0, dx - 1, dy - 1);
	mono->EndDraw ();
	
	GetBitmapBits (mono->ReturnPrivatePixmapInfo ()->hbitmap, bytes, and_bits);
	
	/*
	 *	The relationship between the foreground, background and
	 *	the result on the screen are showed below :
	 *
	 *	B        F o r e g r o u n d
	 *	a         +-------+-------+
	 *	c         | white | black |
	 *	k +-------+-------+-------+
	 *	g | white |   -   :  res  |
	 *	r +-------+ - - - + - - - +
	 *	o | black | white : black |
	 *	u +-------+-------+-------+
	 *	n
	 *	d
	 *
	 *	Please don't use a black foreground on a white background.
	 *	This might be used for some extension in the future !
	 *
	 *	CAUTION : Black is considered to be "0" and white is "1".
	 *
	 *
	 *	The Windows cursor object is quite strange :
	 *
	 *	 xor+---+---+
	 * and  | 0 | 1 |
	 *	+---+---+---+
	 *	| 0 | B | W |
	 *  +---+---+---+
	 *  | 1 | - | X |
	 *  +---+---+---+
	 *
	 *	Windows transformation may be expressed as :
	 *
	 *	dst := (xor & (~and)) | ((src & and) ^ xor);
	 *
	 *	whereas OPaC's transformation is :
	 *
	 *	dst := (fore & (~back)) | (src & fore [& back]);
	 *	
	 *	The "[& back]" term is optional, since currently
	 *	"back => white" always implies "fore => white" !
	 *
	 *	If you think very hard, you will find out that you
	 *	can convert from one format to the other using the
	 *	following expression (you need not check it, since
	 *	it works) :
	 *
	 *	and := fore & back;
	 *	xor := fore & (~back);
	 */
	
	Card32 i = bytes;
	
	while (i--) {
		Card8 b = and_bits[i];
		Card8 f = xor_bits[i];
		
		and_bits[i] = f & b;
		xor_bits[i] = f & ~b;
	}
	
	this->cursor->hcursor = CreateCursor ( OPaC_GrWindow::win_app,
										   (int)(hotx), (int)(hoty), mx, my,
										   and_bits, xor_bits );
	
	//
	//	NB :
	//
	//	The monochrome Grail Pixmap will be deleted automatically when its use
	//	counter will reach zero.
	//
	
	OP_FREE_FROM_ZONE (OPaC_temp_zone, and_bits);
	OP_FREE_FROM_ZONE (OPaC_temp_zone, xor_bits);
	
	this->foreground   = fg;
	this->background   = bg;
	this->x_foreground = (Int16)(xf);
	this->y_foreground = (Int16)(yf);
	this->x_background = (Int16)(xb);
	this->y_background = (Int16)(yb);
	this->dx = (Int16)(dx);
	this->dy = (Int16)(dy);
	this->hot_x = (Int16)(hotx);
	this->hot_y = (Int16)(hoty);
}



/*
 *	Make the cursor visible on screen.
 */

void
OPGrailCursor::Activate () const
{
	OP_METHOD (Activate, 0);
	
	HCURSOR cursor = (this->cursor && this->cursor->hcursor)
					 ? this->cursor->hcursor
					 : LoadCursor (NULL, IDC_ARROW);
	
	SetCursor (cursor);
	
	global_OPaC_cursor = cursor;
}



/*
 *	Switch back to the default cursor.
 */

void
OPGrailCursor::ResetToNormal () const
{
	OP_METHOD (ResetToNormal, 0);
	
	HCURSOR cursor = LoadCursor (NULL, IDC_ARROW);
	
	SetCursor (cursor);
	
	global_OPaC_cursor = cursor;
}



/*
 *	Remove any cursor from the screen.
 */

void
OPGrailCursor::Hide () const
{
	OP_METHOD (Hide, 0);
	SetCursor (0);
}

