/*
 *	app.cxx
 *
 *	Application class implementation. It is the application's responsibility to
 *	dispatch events to the responders.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/app.h>
#include <OPaC/event.h>
#include <OPaC/window.h>
#include <OPaC/new.h>
#include <OPaC/pubmacro.h>
#include <OPaC/notmacro.h>

/****************************************************************/

OP_IMPLEMENTATION (app.h);
OP_IMPLEMENT_CLASS (OPApp);

/****************************************************************/

OP_PUBLISH_FUNCTION (OPApp, "WindowList",   ReturnWindowList,   "List of windows",   OPObject, OPList);
OP_PUBLISH_FUNCTION (OPApp, "WindowNumber", ReturnWindowNum,    "Number of windows", Card32, Card32);
OP_PUBLISH_FUNCTION (OPApp, "ActiveWindow", ReturnActiveWindow, "Active window", OPObject, OPWindow);

/****************************************************************/

OP_NOTIFY_EVENT (OPApp, WindowChanged);
OP_NOTIFY_EVENT (OPApp, ActiveWindowChanged);

OP_NOTIFY_CHANGE (OPApp, WindowList, WindowChanged);
OP_NOTIFY_CHANGE (OPApp, WindowNumber, WindowChanged);
OP_NOTIFY_CHANGE (OPApp, ActiveWindow, ActiveWindowChanged);

/****************************************************************/

/*
 *	Some global variables are needed by the application
 *	class :
 */

OPAppRef OPApp::current_app;
Count    OPApp::current_num   = 0;
Address  OPApp::stack_base    = 0;
Bool     OPApp::is_growing_up = FALSE;

/****************************************************************/

extern void* debugger_ptr;
extern void  debuggerAboutMemory (const void*);
       void* debugger_ptr = 0;

/****************************************************************/

/*
 *	Return TRUE if the stack happens to grow from lower
 *	memory addresses to higher ones. It is easy and quite
 *	straightforward to find out...
 */

static Bool
AnalyseStackGrowth (Address base)
{
	OP_FUNCTION (AnalyseStackGrowth, 1);
	OP_ARG (1, Address, base);
	
	Address current = (Address)(& base);
	return (current > base) ? TRUE : FALSE;
}

/****************************************************************/

/*
 *	Initialise an application instance. This will also
 *	setup the event handler.
 */

OPApp::OPApp ()
{
	OP_METHOD (OPApp, 0);
	OP_ASSERT (OPApp::current_num == 0);
	
	OPEvent::Initialise ();
	
	this->is_building_enabled = TRUE;
	
	this->window_list = OP_NEW_FROM_ZONE (this->ReturnMemZone (), OPList);
	this->window_num  = 0;
	this->first_tick  = 0;
	this->last_tick   = 0;

	this->SetName ("Application");
}



/*
 *	Destroy the application object. This should provide a
 *	complete clean-up and might compute some statistics
 *	for profiling.
 */

OPApp::~OPApp ()
{
	OP_METHOD (~OPApp, 0);
	
	TickResponder* act  = this->first_tick;
	TickResponder* next = 0;
	
	//	Remove every remaining tick responder from the waiting list.
	
	while (act) {
		next = act->next;
		DELETE_FROM_ZONE (OPaC_sys_zone, act);
		act  = next;
	}
	
	//	Terminate the event handler too.
	
	OPEvent::Kill ();
}



/*
 *	If more initialization are needed, they would be
 *	performed here.
 */

void
OPApp::Initialise (Bool enable_build)
{
	OP_METHOD (Initialise, 1);
	OP_ARG (1, Bool, enable_build);
	
//?	OPApp::stack_base = (Address)(& enable_build);
	this->is_building_enabled = enable_build;
	
	//	Create the builder object, if it exists and if building is enabled.
	
	if (enable_build) {
		this->SetBuilder (OP_DYNAMIC_NEW ("OPBuilder", OPObject));
	}
}


/*
 *	Access the global builder.
 */

Bool
OPApp::SetBuilder (OPObjectIn builder)
{
	OP_METHOD (SetBuilder, 1);
	OP_ARG (1, OPObjectIn, builder);
	
	this->builder = builder;
	
	return TRUE;
}

Bool
OPApp::GetBuilder (OPObjectOut builder) const
{
	OP_METHOD (GetBuilder, 1);
	OP_ARG (1, OPObjectOut, builder);
	
	builder = this->builder;
	
	return TRUE;
}

const OPObjectRef&
OPApp::ReturnBuilder () const
{
	OP_METHOD (ReturnBuilder, 0);
	return this->builder;
}

Bool
OPApp::IsBuildEnabled () const
{
	OP_METHOD (IsBuildEnabled, 0);
	return this->builder.IsValid ();
}



/*
 *	Handle the active window.
 */

void
OPApp::NotifyActiveWindow (OPWindowIn window)
{
	OP_METHOD (NotifyActiveWindow, 1);
	OP_ARG (1, OPWindowIn, window);
	
	this->active_window = window;
	this->Report (notify_ActiveWindowChanged);
}

const OPWindowRef&
OPApp::ReturnActiveWindow () const
{
	OP_METHOD (ReturnActiveWindow, 0);
	
	return this->active_window;
}


/*
 *	Storing an application is not possible.
 */

void
OPApp::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (SaveInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	OP_ABORT ("Trying to store Application");
}



/*
 *	Restoring an application is not possible.
 */

void
OPApp::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	OP_ABORT ("Trying to restore Application");
}



/*
 *
 */

void
OPApp::Delete ()
{
	OP_METHOD (Delete, 0);
	
	super::Delete ();
	//	NIY
}



/*
 *
 */

Count
OPApp::TotalCount () const
{
	OP_METHOD (TotalCount, 0);
	
	//	NIY
	
	return super::TotalCount ();
}



/*
 *
 */

Bool
OPApp::Get (Count index, OPObjectOut object)
{
	OP_METHOD (Get, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPObjectOut, object);
	
	//	NIY
	
	return super::Get (index, object);
}



/*
 *
 */

Bool
OPApp::SpecifyExtraction (OPArrayIn x)
{
	OP_METHOD (SpecifyExtraction, 1);
	OP_ARG (1, OPArrayIn, x);
	
	//	NIY
	
	return super::SpecifyExtraction (x);
}



/*
 *	The main event loop is entered after everything has
 *	been set up. This will dispatch the events, till it
 *	gets a "Quit" event.
 */

void
OPApp::MainEventLoop ()
{
	OP_METHOD (MainEventLoop, 0);
	
	const OPEventRef& event = OPEvent::CurrentEvent ();
	
	//	Wait... Dispatch... Wait... Dispatch... Wait...
	//	That's an easy job :-) !
	
	while (event->WaitForEvent ()) {
		
		OPObject::NewReportID ();
		
		if ( (event->ReturnType () == OP_EVENT_KEY_DOWN)
		  && (event->ReturnKey ().type == OP_KEY_FUNCTION)
		  && (event->ReturnKey ().code == 10) ){
			
			//	Developper pressed F10 and set debugger_ptr by some special
			//	means. This will display information about the memory.
			
			debuggerAboutMemory (debugger_ptr);
		}

		event->DispatchEvent ();
	}
}



/*
 *	A tick event has been recieved. "ms" have elapsed
 *	since last tick. This will send events to every
 *	registered responder.
 */

void
OPApp::HandleTick (Count ms)
{
	OP_METHOD (HandleTick, 1);
	OP_ARG (1, Count, ms);
	
	if (this->first_tick) {
		this->first_tick->remaining -= ms;
	}
	
	while ( (this->first_tick)
	     && (this->first_tick->remaining <= 0) ) {
		
		TickResponder* ticker = this->first_tick;
		TickResponder* next   = ticker->next;
		
		this->first_tick = next;
		(next ? next->prev : this->last_tick) = 0;
		
		if (next) next->remaining += ticker->remaining;
		
		ticker->responder->Tick ();
		
		DELETE_FROM_ZONE (OPaC_sys_zone, ticker);
	}
}



/*
 *	Register a window - it might not be visible nor fully
 *	initialised at this point.
 */

void
OPApp::InsertWindow (OPWindowIn window)
{
	OP_METHOD (InsertWindow, 1);
	OP_ARG (1, OPWindowIn, window);
	
	this->window_list->Insert (window);
	this->window_num++;
	this->Report (notify_WindowChanged, OP_REPORT_NEW);
}



/*
 *	A window is no longer available for the application.
 *	This will just remove the window from the app's list.
 */

void
OPApp::RemoveWindow (OPWindowIn window)
{
	OP_METHOD (RemoveWindow, 1);
	OP_ARG (1, OPWindowIn, window);
	
	if (window.IsValid ()) {
		OP_ASSERT (this->window_list->Remove (window));
		this->window_num--;
		this->Report (notify_WindowChanged, OP_REPORT_NEW);
	}
}



/*
 *	Return the nth registered window.
 */

Bool
OPApp::GetWindow (Count index, OPWindowOut window)
{
	OP_METHOD (FindWindow, 2);
	OP_ARG (1, Count, index);
	OP_ARG (3, OPWindowOut, window);
	
	OPObjectRef	obj = 0;
	
	if (this->window_list->Get (index, obj)) {
		window = obj;
		return TRUE;
	}
	
	window = 0;
	return FALSE;
}



/*
 *	Register a responder which we will have to send a
 *	tick event after "ticks" milliseconds.
 */

void
OPApp::InsertTickResponder (OPResponderIn responder, Card32 ticks)
{
	OP_METHOD (InsertTickResponder, 2);
	OP_ARG (1, OPResponderIn, responder);
	OP_ARG (2, Card32, ticks);
	
	TickResponder* ticker  = NEW_FROM_ZONE (OPaC_sys_zone) TickResponder;
	TickResponder* current = this->first_tick;
	
	while (current && (current->remaining <= (Int32)(ticks))) {
		ticks  -= current->remaining;
		current = current->next;
	}
	
	ticker->next      = current;
	ticker->prev      = (current ? current->prev : this->last_tick);
	ticker->remaining = ticks;
	ticker->responder = responder;
	
	(current ? current->prev : this->last_tick) = ticker;
	(ticker->prev ? ticker->prev->next : this->first_tick) = ticker;
	
	if (current) {
		current->remaining -= ticks;
	}
}



/*
 *	A responder ceased to exist : we have to remove any
 *	reference to it in the tick list or else... :-(
 */

void
OPApp::RemoveTickResponder (OPResponderIn responder)
{
	OP_METHOD (RemoveTickResponder, 1);
	OP_ARG (1, OPResponderIn, responder);
	
	TickResponder* act  = this->first_tick;
	TickResponder* next = 0;
	
	while (act) {
		
		next = act->next;
		
		if (act->responder == responder) {
			
			(act->next ? act->next->prev : this->last_tick)  = act->prev;
			(act->prev ? act->prev->next : this->first_tick) = act->next;
			
			if (act->next) act->next->remaining += act->remaining;
			
			DELETE_FROM_ZONE (OPaC_sys_zone, act);
		}
		
		act  = next;
	}
}



/*
 *	Initialise the application class.
 */

void
OPApp::Initialise (OPAppIn app)
{
	OP_MEMBER (Initialise, 1);
	OP_ARG (1, OPAppIn, app);
	OP_ASSERT (OPApp::current_num == 0);
	
	if (OPApp::stack_base == 0) {
		int any_value = 0;
		
		OPApp::stack_base    = (Address)(& any_value);
		OPApp::is_growing_up = AnalyseStackGrowth (OPApp::stack_base);
	}
	
	OPApp::current_app = app;
	OPApp::current_num++;
}



/*
 *	Kill the application class.
 */

void
OPApp::Kill ()
{
	OP_MEMBER (Kill, 0);
	OP_ASSERT (OPApp::current_num == 1);
	
	OPWindowRef window = 0;
	Count       index  = 0;
	
	while (OPApp::current_app->GetWindow (index++, window)) {
		window->Delete ();
	}
	
	if (OPApp::current_app->window_list.IsValid ()) {
		OPApp::current_app->window_list->RemoveAll ();
	}
	
	OPApp::current_app->Delete ();
	OPApp::current_app->MarkAsDeadButOne (OPApp::current_app);
	
	OPApp::current_app = 0;
	OPApp::current_num--;
}



/*
 *	Return TRUE if building and modifying the user
 *	interface is allowed.
 */

Bool
OPApp::IsBuildingEnabled () const
{
	OP_METHOD (IsBuildingEnabled, 0);
	return this->is_building_enabled;
}



/*
 *	Return the current application instance. This is very
 *	similar to NeXTSTEP's "NXApp" global variable...
 */

const OPAppRef&
OPApp::CurrentApp ()
{
	OP_METHOD (CurrentApp, 0);
	
	return OPApp::current_app;
}



/*
 *	Check to see whether the pointer points into the stack
 *	or somewhere else. This depends on the current stack
 *	usage to find out the top of the stack...
 *
 *	Return TRUE if somewhere in the stack.
 */

Bool
OPApp::IsOnStack (const void* ptr)
{
	OP_METHOD (IsOnStack, 1);
	OP_ARG (1, const void*, ptr);
	
	return OPApp::IsOnStack ((Address)(ptr));
}

Bool
OPApp::IsOnStack (Address ptr)
{
	OP_METHOD (IsOnStack, 1);
	OP_ARG (1, Address, ptr);
	
	if (OPApp::stack_base == 0) {
		
		//	In fact, we don't know yet if the address provided by the
		//	caller resides in the stack... Suppose it does not, since
		//	up to this point, only the OPaC library should have made
		//	calls to "IsOnStack".
		
		return FALSE;
	}
	
	Address test = (Address)(ptr);
	Address top  = (Address)(&test);
	
	if (OPApp::is_growing_up) {
		OP_ASSERT (top > OPApp::stack_base);
		if ( (test < OPApp::stack_base)
		  || (test > top) )
			return FALSE;
	} else {
		OP_ASSERT (top < OPApp::stack_base);
		if ( (test > OPApp::stack_base)
		  || (test < top) )
			return FALSE;
	}
	
	return TRUE;
}




const OPListRef&
OPApp::ReturnWindowList () const
{
	OP_METHOD (ReturnWindowList, 0);
	return this->window_list;
}

Count
OPApp::ReturnWindowNum () const
{
	OP_METHOD (ReturnWindowNum, 0);
	return this->window_num;
}
