/*
 *	atom.cxx
 *
 *	OPaC Atom implementation. The OPAtom class maintains a table of allocated
 *	atoms and offers services to create and analyse atoms, using the reference
 *	class.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/atom.h>
#include <OPaC/string.h>
#include <OPaC/new.h>
#include <OPaC/values.h>

/****************************************************************/

OP_IMPLEMENTATION (atom.h);
OP_SPECIAL_CLASS (OPAtom);

/****************************************************************/

/*
 *	Global variables for the OPaC Atom class. They contain
 *	a hash table used to store the atoms.
 */

OPAtom*    OPAtom::table[OP_ATOM_HASH] = { 0 /* ... */ };
OPMemZone* OPAtom::zone;

/****************************************************************/

/*
 *	Initialise an atom. This does not insert the atom into
 *	the table; it just clears its fields.
 */

OPAtom::OPAtom ()
{
	OP_METHOD (OPAtom, 0);
	OP_ASSERT (OPAtom::zone);
	
	this->self     = this;
	this->next     = 0;
	this->text     = 0;
	this->mem_zone = OPAtom::zone;
}


/*
 *	Destroy an atom. This should never happen, unless it
 *	has not been inserted into the table.
 */

OPAtom::~OPAtom ()
{
	OP_METHOD (~OPAtom, 0);
	OP_ASSERT (this->text == 0);
}



/*
 *	Store the text representing the atom. This will be
 *	called by operator << through OPStorage::StoreAtom
 *	method...
 */

void
OPAtom::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	OP_ABORT ("Cannot store that way !");
}



/*
 *	The restore method will never be called, since OPAtom
 *	is a special class. The restoration has to be handled
 *	by operator >> directly !
 */

void
OPAtom::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	OP_ABORT ("Cannot restore that way !");
}



/*
 *	Allocate room for some text from the ATOM zone. This
 *	is useful to keep atoms tightly together in virtual
 *	memory systems.
 */

LatinChar*
OPAtom::AllocText (Count len)
{
	OP_METHOD (AllocText, 1);
	OP_ARG (1, Count, len);
	
	LatinChar* ptr = OP_ALLOC_ARRAY_FROM_ZONE (OPAtom::zone, LatinChar, len);
	
	return ptr;
}



/*
 *	Define the atom's name as a string object.
 */

void
OPAtom::SetName (OPStringIn name)
{
	OP_METHOD (SetName, 1);
	OP_ARG (1, OPStringIn, name);
	
	Count      len = name->CountCharacters () + 1;
	LatinChar* ptr = this->AllocText (len+1);
	
	//	Extract text from the string, including trailing '\0' !
	
	name->GetStringCopy (ptr, len); ptr[len] = 0;
	this->SetText (ptr);
}



/*
 *	Define the atom's name as an ISO Latin-1 character
 *	string.
 */

void
OPAtom::SetName (const LatinChar* name)
{
	OP_METHOD (SetName, 1);
	OP_ARG (1, const LatinChar*, name);
	OP_ASSERT (name && name[0]);
	
	Count            len = 0;
	LatinChar*       ptr = 0;
	const LatinChar* tmp = name;
	
	while (*tmp++) /* nothing */;
	
	//	Copy the text including the trailing '\0'...
	
	len = (Count)(tmp - name);
	ptr = this->AllocText (len);
	
	while (len--) ptr[len] = name[len];
	
	this->SetText (ptr);
}



/*
 *	Define the atom's name as a Unicode character string.
 *	This will first convert the string to ISO Latin-1.
 */

void
OPAtom::SetName (const UnicodeChar* name)
{
	OP_METHOD (SetName, 1);
	OP_ARG (1, const UnicodeChar*, name);
	OP_ASSERT (name && name[0]);
	
	//	Not quite optimal : we use a string object to convert from Unicode to
	//	the ISO Latin-1 character set.
	
	OPStringRef unicode = name;
	this->SetName (unicode);
}



/*
 *	Define the text of the current atom. This will insert
 *	the atom into the table. The text will be used as is
 *	and has to have been allocated using "AllocText".
 *	
 *	This is only possible if the atom hadn't been inserted
 *	before.
 */

void
OPAtom::SetText (LatinChar* name)
{
	OP_METHOD (SetText, 1);
	OP_ARG (1, LatinChar*, name);
	OP_ASSERT (this->text == 0);
	
	//	This will never be called if the atom already exists. Therefore, we need
	//	not look into the data base first.
	
	Count    hash = OPAtom::Hash (name, OP_ATOM_HASH);
	OPAtom*& atom = this->table[hash];
	
	this->text = name;
	this->next = atom;
	
	atom = this;
}



/*
 *	Return the atom's name. This will allocate a string
 *	object and return a reference to it.
 */

OPStringRef
OPAtom::ReturnString () const
{
	OP_METHOD (ReturnString, 0);
	OP_ASSERT (this->text != 0);
	
	//	This will allocate a new OPaC string and copy the text to the string.
	//	The memory zone used for this is the temporary one.
	
	OPStringRef name = this->text;
	return name;
}

Bool
OPAtom::SetString (OPStringIn string)
{
	OP_METHOD (SetString, 1);
	OP_ARG (1, OPStringIn, string);
	
	this->SetName (string);
	return TRUE;
}



/*
 *	Return the pointer to the text, either as Latin or as
 *	Unicode representation. Currently, atoms are only stored
 *	as Latin encoding...
 */

const LatinChar*
OPAtom::ReturnQuickLatin () const
{
	OP_METHOD (ReturnQuickLatin, 0);
	return this->text;
}

const UnicodeChar*
OPAtom::ReturnQuickUnicode () const
{
	OP_METHOD (ReturnQuickUnicode, 0);
	return 0;
}


/*
 *	Return the hash value of the current atom. Return 0 if
 *	the atom is not defined. A null "num" may not be used.
 */

Count
OPAtom::ComputeHashValue (Count num) const
{
	OP_METHOD (ComputeHashValue, 1);
	OP_ARG (1, Count, num);
	OP_ASSERT (num);
	
	return OPAtom::Hash (this->text, num);
}



/*
 *	Compare the atom with another one. If the object is not
 *	of the OPAtom type, then the default comparison method
 *	will be applied.
 */

int
OPAtom::CompareWith (OPObjectIn object) const
{
	OP_METHOD (CompareWith, 1);
	OP_ARG (1, OPObjectIn, object);
	
	if (object->IsKindOf (OP_RUNTIME (OPAtom))) {
		
		OPAtomRef atom = object;
		
		const LatinChar* text_a = this->text;
		const LatinChar* text_b = atom->text;
		
		OP_ASSERT (text_a);
		OP_ASSERT (text_b);
		
		//	Compare character by character. The comparison are made using the
		//	character's ISO Latin-1 codes.
		
		for (;;) {
			
			if ( (*text_a == 0)
			  || (*text_b == 0)
			  || (*text_a != *text_b) )
				break;
			
			text_a++;
			text_b++;
		}
		
		if (*text_a < *text_b)
			return OP_COMP_THIS_BEFORE;
		
		if (*text_a > *text_b)
			return OP_COMP_THIS_AFTER;
		
		return OP_COMP_THIS_SAME;
	}
	
	return super::CompareWith (object);
}

/****************************************************************/


/*
 *	Initialise the OPaC Atom DataBase Manager: this mainly
 *	creates a memory zone for further allocations.
 */

void
OPAtom::Initialise ()
{
	OP_MEMBER (Initialise, 0);
	OP_ASSERT (OPAtom::zone == 0);
	
	//	Initialise the memory zone associated to the OPaC atoms.
	
	OPAtom::zone = NEW (OPMemZone);
	OPAtom::zone->Initialise ("AtomZone");
}



/*
 *	Free the atoms... This will free memory for every
 *	existing atom in the OPaC world.
 */

void
OPAtom::Kill ()
{
	OP_MEMBER (Kill, 0);
	OP_ASSERT (OPAtom::zone);
	
	Count hash;
	
	for (hash = 0; hash < OP_ATOM_HASH; hash++) {
		
		OPAtom* atom = OPAtom::table[hash];
		OPAtom* next = 0;
		
		while (atom) {
			
			next = atom->next;

			OP_ASSERT (atom->ReturnMemZone () == OPAtom::zone);
			
			atom->MarkAsDead ();
			atom->Delete ();
			
			if (atom->text) {
				OP_FREE_FROM_ZONE (OPAtom::zone, atom->text);
				atom->text = 0;
			}
			
			DELETE_FROM_ZONE (OPAtom::zone, atom);
			atom = next;
		}
		
		OPAtom::table[hash] = 0;
	}
	
	//	Now kill the atom zone. That's all folks !
	
	DELETE (OPAtom::zone);
	OPAtom::zone = 0;
}



/*
 *	Allocates a new atom, using the atom specific memory
 *	zone.
 */

OPAtom*
OPAtom::NewAtom ()
{
	OP_MEMBER (NewAtom, 0);
	OP_ASSERT (OPAtom::zone);
	
	return (OPAtom*)(OPClassMan::New (OP_RUNTIME (OPAtom),
									  OPAtom::zone,
									  __LINE__, __FILE__));
}


/*
 *	Find an atom which matches the specified name. If none
 *	can be found, return 0.
 */

OPAtom*
OPAtom::FindAtom (OPStringIn name)
{
	OP_MEMBER (FindAtom, 1);
	OP_ARG (1, OPStringIn, name);
	
	if (name.IsValid ()) {
		OPAtom*   atom = 0;
		Count     len  = name->CountCharacters () + 1;
		LatinChar temp[256];
		
		OP_ASSERT (len < 256);
		
		name->GetStringCopy (temp, len);
		atom = OPAtom::FindAtom (temp);
		
		return atom;
	}
	
	return 0;
}



/*
 *	Find an atom which matches the specified name. If none
 *	can be found, return 0.
 */

OPAtom*
OPAtom::FindAtom (const LatinChar* name)
{
	OP_MEMBER (FindAtom, 1);
	OP_ARG (1, const LatinChar*, name);
	
	if ( (name == 0)
	  || (name[0] == 0) ) {
		return 0;
	}
	
	Count   hash = OPAtom::Hash (name, OP_ATOM_HASH);
	OPAtom* atom = OPAtom::table[hash];
	
	while (atom) {
		
		const LatinChar* text_a = name;
		const LatinChar* text_b = atom->text;
		
		while ( (text_a[0] == text_b[0])
		     && (text_a[0] != 0)
		     && (text_b[0] != 0) ) {
			text_a++;
			text_b++;
		}
		
		if (text_a[0] == text_b[0])
			break;
		
		atom = atom->next;
	}
	
	return atom;
}



/*
 *	Find an atom which matches the specified name. If none
 *	can be found, return 0.
 */

OPAtom*
OPAtom::FindAtom (const UnicodeChar* name)
{
	OP_MEMBER (FindAtom, 1);
	OP_ARG (1, const UnicodeChar*, name);
	
	if ( (name == 0)
	  || (name[0] == 0) ) {
		return 0;
	}
	
	//	Not quite optimal : we use a string object to convert from Unicode to
	//	the ISO Latin-1 character set.
	
	OPStringRef unicode = name;
	OPAtom*     atom    = 0;
	
	atom = OPAtom::FindAtom (unicode);
	
	return atom;
}



/*
 *	Return the hash value of the current atom. Return 0 if
 *	the atom is not defined. A null "num" may not be used.
 */

Count
OPAtom::Hash (const LatinChar* text, Count num)
{
	OP_MEMBER (Hash, 2);
	OP_ARG (1, const LatinChar*, text);
	OP_ARG (2, Count, num);
	
	if (text == 0) {
		return 0;
	}
	
	Card32 hash = 0;
	Card32 seed = 0;
	Card32 temp = 0;
	
	for (;;) {
		
		temp = (Card32)(*text++);
		
		if (temp == 0)
			break;
		
		seed += 13;
		hash += temp ^ seed;
	}
	
	return hash % num;
}


/****************************************************************/

/*
 *	Create an atom reference (kind of construction).
 *	This uses the "operator =" to do the real job.
 */

OPAtomRef::OPAtomRef (OPStringIn string)		 { (*this) = string; }
OPAtomRef::OPAtomRef (const LatinChar* string)   { (*this) = string; }
OPAtomRef::OPAtomRef (const UnicodeChar* string) { (*this) = string; }

/****************************************************************/

OPStorageIn
operator >> (OPStorageIn storage, OPAtomOut atom)
{
	storage->RestoreAtom (atom);
	return storage;
}

OPStorageIn operator >> (OPStorageIn storage, OPAtomArrayOut a)
{
	OP_ABORT ("NIY");
	return storage;
}

/****************************************************************/

/*
 *	Atom creation from an anonymous object must handle the
 *	conversion for the known types.
 */

OPAtomRef::OPAtomRef (OPObjectIn x) : OPTextRef ()
{
	if (x.IsValid ()) {
		if (x->IsKindOf (OP_RUNTIME (OPAtom))) {
			(*this) = (OPAtom*)x.ReturnObject ();
		} else if (x->IsKindOf (OP_RUNTIME (OPString))) {
			OPStringRef s = (OPString*)x.ReturnObject ();
			(*this) = s;
		} else if (x->IsKindOf (OP_RUNTIME (OPText))) {
			OPTextRef   b = (OPText*)x.ReturnObject ();
			OPStringRef s = b->ReturnString ();
			(*this) = s;
		} else if (x->IsKindOf (OP_RUNTIME (OPValue))) {
			OPValueRef  v = (OPValue*)x.ReturnObject ();
			OPAtomRef a;
			v->ConvertToAtom (a);
			(*this) = a;
		} else {
			(*this) = (OPObject*)(0);
		}
	} else {
		this->_Set (OPObjectNull);
	}
}

OPAtomRef::OPAtomRef (OPAtomIn x) : OPTextRef (x)
{
}

OPAtomRef::OPAtomRef (OPAtom* x) : OPTextRef (x)
{
}

OPAtomRef&
OPAtomRef::operator = (OPObjectIn x)
{
	if (x.IsValid ()) {
		if (x->IsKindOf (OP_RUNTIME (OPAtom))) {
			(*this) = (OPAtom*)x.ReturnObject ();
		} else if (x->IsKindOf (OP_RUNTIME (OPString))) {
			OPStringRef s = (OPString*)x.ReturnObject ();
			(*this) = s;
		} else if (x->IsKindOf (OP_RUNTIME (OPText))) {
			OPTextRef   b = (OPText*)x.ReturnObject ();
			OPStringRef s = b->ReturnString ();
			(*this) = s;
		} else if (x->IsKindOf (OP_RUNTIME (OPValue))) {
			OPValueRef  v = (OPValue*)x.ReturnObject ();
			OPAtomRef a;
			v->ConvertToAtom (a);
			(*this) = a;
		} else {
			(*this) = (OPObject*)(0);
		}
	} else {
		this->_Set (OPObjectNull);
	}
	return (*this);
}

OPAtomRef&
OPAtomRef::operator = (OPAtomIn x)
{
	this->_Set (x);
	return (*this);
}

OPAtomRef&
OPAtomRef::operator = (OPAtom* x)
{
	this->_Set (x);
	return (*this);
}

/****************************************************************/

/*
 *	An atom reference will not always allocate a new atom
 *	instance, when initialised. It may also use an existing
 *	one (it is the atom's purpose to share texts !)
 */

OPAtomRef&
OPAtomRef::operator = (OPStringIn string)
{
	OP_OPERATOR (=, 1);
	OP_ARG (1, OPStringIn, string);
	
	OPAtom* atom = 0;
	
	if (string.IsValid ()) {
		atom = OPAtom::FindAtom (string);
		
		if (atom == 0) {
			atom = OPAtom::NewAtom ();
			atom->SetName (string);
		}
	}
	
	this->_Set (atom);
	return *this;
}


OPAtomRef&
OPAtomRef::operator = (const LatinChar* string)
{
	OP_OPERATOR (=, 1);
	OP_ARG (1, const LatinChar*, string);
	
	OPAtom* atom = 0;
	
	if (string && string[0]) {
		
		atom = OPAtom::FindAtom (string);
		
		if (atom == 0) {
			atom = OPAtom::NewAtom ();
			atom->SetName (string);
		}
	}
	
	this->_Set (atom);
	return *this;
}

OPAtomRef&
OPAtomRef::operator = (const UnicodeChar* string)
{
	OP_OPERATOR (=, 1);
	OP_ARG (1, const UnicodeChar*, string);
	
	OPAtom* atom = 0;
	
	atom = OPAtom::FindAtom (string);
	
	if (atom == 0) {
		atom = OPAtom::NewAtom ();
		atom->SetName (string);
	}
	
	this->_Set (atom);
	return *this;
}
	
/****************************************************************/

Bool
OPAtomRef::operator == (OPStringIn name) const
{
	OP_OPERATOR (==, 1);
	OP_ARG (1, OPStringIn, name);
	
	return name == ((OPAtom*)(this->object))->text;
}


Bool
OPAtomRef::operator == (const LatinChar* text) const
{
	OP_OPERATOR (==, 1);
	OP_ARG (1, const LatinChar*, text);
	
	OPStringRef name = (*this)->ReturnString ();
	return name == text;
}


Bool
OPAtomRef::operator == (const UnicodeChar* text) const
{
	OP_OPERATOR (==, 2);
	OP_ARG (1, const UnicodeChar*, text);
	
	OPStringRef name = (*this)->ReturnString ();
	return name == text;
}

