/*
 *	convert.cxx
 *
 *	The convert class provides static members to convert any data types to
 *	a string representation, maybe using a special format specifier.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <string.h>
#include <stdio.h>

#include <OPaC/convert.h>
#include <OPaC/new.h>
#include <OPaC/string.h>

OP_IMPLEMENTATION (convert.h);

/*
 *	Convert a boolean value to either TRUE or FALSE if no
 *	valid format is specified.
 *	
 *	It is possible, eg, to specify one of the following
 *	format strings :
 *	
 *	 - "ENUM:#yes#no"
 *	 - "ENUM:+Ok doc !+That's a pity..."
 *	
 *	The syntax is "ENUM<separator>True<separator>False".
 */

void
OPConvert::BoolToString (Bool value, OPStringIn fmt, OPStringOut data)
{
	OP_FUNCTION (OPConvert::BoolToString, 3);
	OP_ARG (1, Bool, value);
	OP_ARG (2, OPStringIn, fmt);
	OP_ARG (3, OPStringOut, data);
	
	if (data.IsInvalid ()) {
		data = OP_NEW (OPString);
	}
	
	Card32    v_card = (value) ? 0 : 1;
	Count     length = (fmt.IsValid ()) ? fmt->CountCharacters () : 0;
	LatinChar buffer[6];

	if (length >= 4) {
		fmt->ExtractStringCopy (buffer, 0, 5);
		if (strcmp (buffer, "ENUM:") == 0) {
			OPConvert::CardToString (v_card, fmt, data);
			return;
		}
	}
	
	data = value ? "TRUE" : "FALSE";
}



/*
 *	Convert an integer to either an n-digit integer or a
 *	name, specified as an enumeration. The accepted syntax
 *	are as following :
 *	
 *	 - "ENUM:<sep>t0<sep>t1<sep>t2<sep>...<sep>tn"
 *	 - "NUM:[text]%[+| ][-][width];[text]"
 *	
 *	The enumeration format specifier must list each possible
 *	value. If the value falls after the last specified tag,
 *	then the value is output as a number. The value may not
 *	be negative.
 *
 *	The numeric format specifier may have some text embedded
 *	within it. If a "%" has to be used in a text before the
 *	real "%", use "%%". If "+" is specified, a sign will be
 *	inserted anyway; if " " is specified, a minus sign will
 *	be inserted if needed and a space otherwise. If "-" is
 *	specified, then the value will be left aligned within
 *	the available width. The output is in decimal.
 */

void
OPConvert::IntToString (Int32 value, OPStringIn fmt, OPStringOut data)
{
	OP_FUNCTION (OPConvert::IntToString, 3);
	OP_ARG (1, Int32, value);
	OP_ARG (2, OPStringIn, fmt);
	OP_ARG (3, OPStringOut, data);
	
	OPConvert::IntToString ((Int32)(value), fmt, "%ld", data);
}



/*
 *	Convert a cardinal to either an n-digit integer or a
 *	name, specified as an enumeration. The accepted syntax
 *	are as following :
 *	
 *	 - "ENUM:<sep>t0<sep>t1<sep>t2<sep>...<sep>tn"
 *	 - "NUM:[text]%[-][width];[text]"
 *	
 *	The enumeration format specifier must list each possible
 *	value. If the value falls after the last specified tag,
 *	then the value is output as a number.
 *
 *	The numeric format specifier may have some text embedded
 *	within it. If a "%" has to be used in a text before the
 *	real "%", use "%%". If "-" is specified, then the value
 *	will be left aligned within the available width. The
 *	output is in decimal.
 */

void
OPConvert::CardToString (Card32 value, OPStringIn fmt, OPStringOut data)
{
	OP_FUNCTION (OPConvert::CardToString, 3);
	OP_ARG (1, Card32, value);
	OP_ARG (2, OPStringIn, fmt);
	OP_ARG (3, OPStringOut, data);
	
	OPConvert::IntToString ((Int32)(value), fmt, "%lu", data);
}



/*
 *	This member works for both IntToString and CardToString
 *	conversions. Don't use it yourself.
 */

void
OPConvert::IntToString (Int32 value,
						OPStringIn fmt,
						const LatinChar* print,
						OPStringOut data)
{
	OP_FUNCTION (OPConvert::IntToString, 4);
	OP_ARG (1, Card32, value);
	OP_ARG (2, OPStringIn, fmt);
	OP_ARG (3, const LatinChar*, print);
	OP_ARG (4, OPStringOut, data);
	
	if (data.IsInvalid ()) {
		data = OP_NEW (OPString);
	}
	
	LatinChar          buffer[200+1];
	Count              length  = (fmt.IsValid ()) ? fmt->CountCharacters () : 0;
	const LatinChar*   latin   = 0;
	const UnicodeChar* unicode = 0;
	
	data->ClearString ();
	
	if (length) {
		latin   = fmt->AccessLatin ();
		unicode = fmt->AccessUnicode ();
		fmt->ExtractStringCopy (buffer, 0, 200);
	} else {
		latin   = "NUM:%;";
		unicode = 0;
		length  = strlen (latin);
		strcpy (buffer, latin);
	}
	
	//	The value is handled as an enumeration. The syntax is defined as
	//	following : "ENUM<sep>0<sep>1<sep>2<sep>3..."
	
	if (strncmp (buffer, "ENUM:", 5) == 0) {
		
		//	Find the proper Latin string in the specified enumeration.
		
		if (latin) {
			LatinChar sep   = latin[5];
			Count     val   = value;
			Count     index = 6;
			Count     copy  = 0;
			
			while (latin[index]) {
				if (val == 0) {
					while ( (latin[index+copy]) && (latin[index+copy] != sep) ) {
						copy++;
					}
					data->AppendString (latin+index, copy);
					return;
				}
				if (latin[index] == sep) {
					val--;
				}
				index++;
			}
		}
		
		//	Find the proper Unicode string in the specified enumeration.
		
		if (unicode) {
			UnicodeChar sep   = unicode[5];
			Count       val   = value;
			Count       index = 6;
			Count       copy  = 0;
			
			while (unicode[index]) {
				if (val == 0) {
					while ( (unicode[index+copy]) && (unicode[index+copy] != sep) ) {
						copy++;
					}
					data->AppendString (unicode+index, copy);
					return;
				}
				if (unicode[index] == sep) {
					val--;
				}
				index++;
			}
		}

		//	The value had no enumeration equivalent. Darn ! Put the cardinal
		//	into the output string, that's the best we can do.
		
		sprintf (buffer, print, value);
		data->AppendString (buffer);
		return;
	}

	//	Output a formatted numeric value. That may be somewhat compilcated
	//	since we have to extract the text and can't/won't use sprintf as is
	//	because we are not too confident in the caller's syntax.
	
	if (strncmp (buffer, "NUM:", 4) == 0) {
		
		//	Output the numeric value using the Latin specifier.
		
		if (latin) {
			
			LatinChar*       bufptr = buffer + 4;
			const LatinChar* copy   = latin + 4;
			Count            count  = 0;
			
			length -= 4;
			latin  += 4;
			
			while (length) {
				if (latin[0] == '%') {
					if (latin[1] == '%') {
						
						//	If "%%" is found, insert just a "%" and copy the previous
						//	text, if any.
						
						data->AppendString (copy, count+1);
						bufptr += 2;
						latin  += 2;
						length -= 2;
						copy    = latin;
						count   = 0;
						
					} else {
						
						//	If a single "%" has been found, we have to extract just the
						//	format command which can then be used by sprintf.
						
						if (count) {
							data->AppendString (copy, count);
						}
						
						LatinChar  spec[64];
						LatinChar* ptr = spec;
						
						latin++;
						length--;
						*ptr++ = '%';
						
						while ( (length)
						     && ( (*bufptr == '-')
						       || ((*bufptr >= '0') && (*bufptr <= '9'))
						       || (*bufptr == ' ')
						       || (*bufptr == '+') ) ) {
							*ptr++ = *bufptr++;
							latin++;
							length--;
						}
						
						//	If the caller did well his job, then there has to be a ";"
						//	after the specifier. If so skip it, if not, sigh...
						
						if (*latin == ';') {
							latin++;
							length--;
						}
						
						strcpy (ptr, print+1);
						sprintf (buffer, spec, value);
						
						data->AppendString (buffer);
						data->AppendString (latin);
						
						return;
					}
					
				} else {
					count++;
					latin++;
					bufptr++;
					length--;
				}
			}
			
			data->AppendString (copy);
			return;
		}
		
		//	Output the numeric value using the Unicode specifier.
		
		if (unicode) {
			
			LatinChar*         bufptr = buffer + 4;
			const UnicodeChar* copy   = unicode + 4;
			Count              count  = 0;
			
			length  -= 4;
			unicode += 4;
			
			while (length) {
				if (unicode[0] == '%') {
					if (unicode[1] == '%') {
						
						//	If "%%" is found, insert just a "%" and copy the previous
						//	text, if any.
						
						data->AppendString (copy, count+1);
						bufptr  += 2;
						unicode += 2;
						length  -= 2;
						copy     = unicode;
						count    = 0;
						
					} else {
						
						//	If a single "%" has been found, we have to extract just the
						//	format command which can then be used by sprintf.
						
						if (count) {
							data->AppendString (copy, count);
						}
						
						LatinChar  spec[64];
						LatinChar* ptr = spec;
						
						latin++;
						length--;
						*ptr++ = '%';
						
						while ( (length)
						     && ( (*bufptr == '-')
						       || ((*bufptr >= '0') && (*bufptr <= '9'))
						       || (*bufptr == ' ')
						       || (*bufptr == '+') ) ) {
							*ptr++ = *bufptr++;
							unicode++;
							length--;
						}
						
						//	If the caller did well his job, then there has to be a ";"
						//	after the specifier. If so skip it, if not, sigh...
						
						if (*unicode == ';') {
							unicode++;
							length--;
						}
						
						strcpy (ptr, print+1);
						sprintf (buffer, spec, value);
						
						data->AppendString (buffer);
						data->AppendString (unicode);
						
						return;
					}
					
				} else {
					count++;
					unicode++;
					bufptr++;
					length--;
				}
			}
			
			data->AppendString (copy);
			return;
		}
	}

	//	Default : just output the numeric value.
	
	sprintf (buffer, print, value);
	data->AppendString (buffer);
}



/*
 *	Convert a real, floating point, value to some string.
 *	The format is similar to the one used by the C printf
 *	functon :
 *
 *	 - "NUM:[text]%[+| ][-][width][.precision];[text]"
 *	 - "EXP:[text]%[+| ][-][width][.precision];[text]"
 */

void
OPConvert::RealToString (Real value, OPStringIn fmt, OPStringOut data)
{
	OP_FUNCTION (OPConvert::RealToString, 3);
	OP_ARG (1, Real, value);
	OP_ARG (2, OPStringIn, fmt);
	OP_ARG (3, OPStringOut, data);
	
	if (data.IsInvalid ()) {
		data = OP_NEW (OPString);
	}
	
	const LatinChar*   print   = "%lf";
	LatinChar          buffer[200+1];
	Count              length  = (fmt.IsValid ()) ? fmt->CountCharacters () : 0;
	const LatinChar*   latin   = 0;
	const UnicodeChar* unicode = 0;
	
	data->ClearString ();
	
	if (length) {
		latin   = fmt->AccessLatin ();
		unicode = fmt->AccessUnicode ();
		fmt->ExtractStringCopy (buffer, 0, 200);
	} else {
		latin   = "NUM:%;";
		unicode = 0;
		length  = strlen (latin);
		strcpy (buffer, latin);
	}
	
	Count skip = 0;

	//	Output a formatted numeric value. That may be somewhat compilcated
	//	since we have to extract the text and can't/won't use sprintf as is
	//	because we are not too confident in the caller's syntax.
	
	if (strncmp (buffer, "NUM:", 4) == 0) {
		skip  = 4;
		print = "%lf";
	}

	if (strncmp (buffer, "EXP:", 4) == 0) {
		skip  = 4;
		print = "%lE";
	}

	if (skip) {
		
		//	Output the numeric value using the Latin specifier.
		
		if (latin) {
			
			LatinChar*       bufptr = buffer + skip;
			const LatinChar* copy   = latin + skip;
			Count            count  = 0;
			
			length -= skip;
			latin  += skip;
			
			while (length) {
				if (latin[0] == '%') {
					if (latin[1] == '%') {
						
						//	If "%%" is found, insert just a "%" and copy the previous
						//	text, if any.
						
						data->AppendString (copy, count+1);
						bufptr += 2;
						latin  += 2;
						length -= 2;
						copy    = latin;
						count   = 0;
						
					} else {
						
						//	If a single "%" has been found, we have to extract just the
						//	format command which can then be used by sprintf.
						
						if (count) {
							data->AppendString (copy, count);
						}
						
						LatinChar  spec[64];
						LatinChar* ptr = spec;
						
						latin++;
						length--;
						*ptr++ = '%';
						
						while ( (length)
						     && ( (*bufptr == '-')
						       || ((*bufptr >= '0') && (*bufptr <= '9'))
						       || (*bufptr == ' ')
						       || (*bufptr == '.')
						       || (*bufptr == '+') ) ) {
							*ptr++ = *bufptr++;
							latin++;
							length--;
						}
						
						//	If the caller did well his job, then there has to be a ";"
						//	after the specifier. If so skip it, if not, sigh...
						
						if (*latin == ';') {
							latin++;
							length--;
						}
						
						strcpy (ptr, print+1);
						sprintf (buffer, spec, value);
						
						data->AppendString (buffer);
						data->AppendString (latin);
						
						return;
					}
					
				} else {
					count++;
					latin++;
					bufptr++;
					length--;
				}
			}
			
			data->AppendString (copy);
			return;
		}
		
		//	Output the numeric value using the Unicode specifier.
		
		if (unicode) {
			
			LatinChar*         bufptr = buffer + skip;
			const UnicodeChar* copy   = unicode + skip;
			Count              count  = 0;
			
			length  -= skip;
			unicode += skip;
			
			while (length) {
				if (unicode[0] == '%') {
					if (unicode[1] == '%') {
						
						//	If "%%" is found, insert just a "%" and copy the previous
						//	text, if any.
						
						data->AppendString (copy, count+1);
						bufptr  += 2;
						unicode += 2;
						length  -= 2;
						copy     = unicode;
						count    = 0;
						
					} else {
						
						//	If a single "%" has been found, we have to extract just the
						//	format command which can then be used by sprintf.
						
						if (count) {
							data->AppendString (copy, count);
						}
						
						LatinChar  spec[64];
						LatinChar* ptr = spec;
						
						latin++;
						length--;
						*ptr++ = '%';
						
						while ( (length)
						     && ( (*bufptr == '-')
						       || ((*bufptr >= '0') && (*bufptr <= '9'))
						       || (*bufptr == ' ')
						       || (*bufptr == '.')
						       || (*bufptr == '+') ) ) {
							*ptr++ = *bufptr++;
							unicode++;
							length--;
						}
						
						//	If the caller did well his job, then there has to be a ";"
						//	after the specifier. If so skip it, if not, sigh...
						
						if (*unicode == ';') {
							unicode++;
							length--;
						}
						
						strcpy (ptr, print+1);
						sprintf (buffer, spec, value);
						
						data->AppendString (buffer);
						data->AppendString (unicode);
						
						return;
					}
					
				} else {
					count++;
					unicode++;
					bufptr++;
					length--;
				}
			}
			
			data->AppendString (copy);
			return;
		}
	}
	
	//	Default : just output the numeric value.
	
	sprintf (buffer, print, value);
	data->AppendString (buffer);
}



/*
 *	Convert an atom to a string. The format specifier is
 *	currently not used.
 */

void
OPConvert::AtomToString (OPAtomIn value, OPStringIn fmt, OPStringOut data)
{
	OP_FUNCTION (OPConvert::AtomToString, 3);
	OP_ARG (1, OPAtomIn, value);
	OP_ARG (2, OPStringIn, fmt);
	OP_ARG (3, OPStringOut, data);
	
	data = value;
}



/*
 *	Convert a string to a ... string. The format specifier
 *	is currently not used.
 */

void
OPConvert::StringToString (OPStringIn value, OPStringIn fmt, OPStringOut data)
{
	OP_FUNCTION (OPConvert::StringToString, 3);
	OP_ARG (1, OPStringIn, value);
	OP_ARG (2, OPStringIn, fmt);
	OP_ARG (3, OPStringOut, data);
	
	data = value;
}



void
OPConvert::StringToInt (OPStringIn value, Int32Out data)
{
	OP_FUNCTION (OPConvert::StringToInt, 2);
	OP_ARG (1, OPStringIn, value);
	OP_ARG (2, Int32Out, data);
	
	if (value.IsInvalid ()) {
		data = 0;
		return;
	}
	
	LatinChar buffer[40];
	value->GetStringCopy (buffer, 40);
	data = (Int32)(atol (buffer));
}

void
OPConvert::StringToCard (OPStringIn value, Card32Out data)
{
	OP_FUNCTION (OPConvert::StringToCard, 2);
	OP_ARG (1, OPStringIn, value);
	OP_ARG (2, Card32Out, data);
	
	if (value.IsInvalid ()) {
		data = 0;
		return;
	}
	
	LatinChar buffer[40];
	value->GetStringCopy (buffer, 40);
	data = (Card32)(atol (buffer));
}

void
OPConvert::StringToReal (OPStringIn value, RealOut data)
{
	OP_FUNCTION (OPConvert::StringToReal, 2);
	OP_ARG (1, OPStringIn, value);
	OP_ARG (2, RealOut, data);
	
	if (value.IsInvalid ()) {
		data = 0.0;
		return;
	}
	
	LatinChar buffer[40];
	value->GetStringCopy (buffer, 40);
	data = (Real)(atof (buffer));
}

