/*
 *	dialog.cxx
 *
 *	The OPaC Dialog class provides some mean to open and let the user play
 *	with some user interface dialog.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/app.h>
#include <OPaC/archive.h>
#include <OPaC/database.h>
#include <OPaC/dialog.h>
#include <OPaC/new.h>
#include <OPaC/stream.h>
#include <OPaC/window.h>

/****************************************************************/

OP_IMPLEMENTATION (dialog.h);
OP_IMPLEMENT_CLASS (OPDialog);

/****************************************************************/

OPDialogRef OPDialog::head;
OPDialogRef OPDialog::tail;

/****************************************************************/


OPDialog::OPDialog ()
{
	OP_METHOD (OPDialog, 0);

	this->prev = this->tail;
	this->next = 0;

	(this->prev.IsValid () ? this->prev->next : this->head) = this;
	(this->next.IsValid () ? this->next->prev : this->tail) = this;
}


OPDialog::~OPDialog ()
{
	OP_METHOD (~OPDialog, 0);
	OP_ASSERT (this->prev.IsValid () == FALSE);
	OP_ASSERT (this->next.IsValid () == FALSE);
}
	

void
OPDialog::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);

	//	NIY
}

void
OPDialog::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);

	//	NIY
}


void
OPDialog::Delete ()
{
	OP_METHOD (Delete, 0);

	if ( (this->next.IsValid ())
	  || (this->prev.IsValid ())
	  || (this->head == this) ) {
		
		(this->next.IsValid () ? this->next->prev : this->tail) = this->prev;
		(this->prev.IsValid () ? this->prev->next : this->head) = this->next;
		
		this->next = 0;
		this->prev = 0;
	}
	
	if (this->window.IsValid ()) {
		this->window->Delete ();
	}

	this->window = OPWindowNull;
	this->name   = OPAtomNull;
}



/*
 *	Interaction... creates a dialog and handle it. This tries
 *	to open the named dialog from a file, on success :
 *	
 *	  o Setup the archive to read from the file.
 *
 *	  o Tell the archive what data slot has to be made available
 *	    to the objects which will be restored (the dialog is the
 *	    data slot).
 *	
 *	  o Restore the topmost object (the window) which will in
 *	    turn restore every subview, then awake it.
 *	
 *	  o Close the archive and the file.
 *	
 *	And on failure :
 *	
 *	  o Create a new window, then setup the interactive dialog
 *	    builder tools.
 */

Bool
OPDialog::Interaction (OPAtomIn name)
{
	OP_MEMBER (Interaction, 1);
	OP_ARG (1, OPAtomIn, name);
	
	OPAppRef      app  = OPApp::CurrentApp ();
	OPDataSlotRef slot = this;
	
	this->AddField ("Application", app);
	this->AddField ("Data slot", slot);
	
	this->ReturnFieldBase ()->SpecifyExtractAttribute ("FieldName", "FieldData");

	
	OPFileReadRef file    = OP_NEW (OPFileRead);
	OPArchiveRef  archive = OP_NEW (OPArchive);
	OPRuntimeRef  rt;
	
	TRY {
		file->OpenRead (name->ReturnString ());
		
		archive->SetDataSlot (this);
		archive->SetInput (file);
		archive->RestoreObject (this->window, rt);
		archive->FinallyAwakeFromApp (OPApp::CurrentApp ());
		archive->Reset ();
		
		file->Close ();
	}
	CATCH (OPExcStream, exc) {
	}
	END_CATCH;
	
	if (this->window.IsValid () == FALSE) {
		this->window = OP_NEW (OPWindow);
		this->window->SetOrigin (50, 50);
		this->window->SetSize (320, 240);
		this->window->SetStyle (OP_WIN_TITLE | OP_WIN_MOVE_OK | OP_WIN_CLOSE_OK);
		this->window->SetTitle (name->ReturnString ());
		this->window->SetSlot (this);
	}
	
	this->name = name;
	this->window->Open ();
	this->window->Display ();
	this->window->ShowNormal ();
	this->window->SetSlot (slot);
	
	return TRUE;
}


Bool
OPDialog::GetWindow (OPWindowOut window) const
{
	window = this->window;
	return window.IsValid ();
}

Bool
OPDialog::GetName (OPAtomOut name) const
{
	name = this->name;
	return name.IsValid ();
}
