/*
 *	grail.cxx
 *
 *	GRAphical Interface Layer for building graphical interface
 *	elements. This is part should not be modified when the OPaC
 *	system gets adapted to a new target.
 *	
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/grail.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>

#include <OPaC/sd/grail.h>

/****************************************************************/

OP_IMPLEMENTATION (grail.h);
OP_ABSTRACT_CLASS (OPGrailPort);

/****************************************************************/


/*
 *	Initialises a Grail Port and allocate some of the
 *	private structure fields (eg. a region for clipping).
 */

OPGrailPort::OPGrailPort ()
{
	OP_METHOD (OPGrailPort, 0);
	
	this->info        = 0;
	this->application = 0;									//	setup by "InitApp"
	this->color_man   = OPColorMan::ReturnSystemColorMan ();
	this->font_man    = OPFontMan::ReturnSystemFontMan ();
	this->adorner     = 0;									//	lazy allocation...
	
	OP_ASSERT (this->color_man.IsValid ());
	OP_ASSERT (this->font_man.IsValid ());
	
	this->rect_x1 = this->rect_y1 = 0;
	this->rect_x2 = this->rect_y2 = 0;
	this->clip_x1 = this->clip_y1 = 0;
	this->clip_x2 = this->clip_y2 = 0;
	
	this->line_style = OP_LINE_SOLID;			//	default is solid line style
	this->font       = 0;
	this->fore_color = this->color_man->FindColor (0.0, 0.0, 0.0);	//	black
	this->back_color = this->color_man->FindColor (1.0, 1.0, 1.0);	//	white
	
	this->use_transparency  = FALSE;
	this->transparent_color = this->back_color;
	
	this->is_ready   = FALSE;					//	will be initialised by a subclass
	this->is_drawing = FALSE;					//	currently not drawing
	
	OPaC_GrailPort_Initialise (this->info, this->fore_color, this->back_color);
}



/*
 *	Clean up a Grail Port. This will destroy everything
 *	that has been constructed by the constructor (logical,
 *	huh ?)
 */

OPGrailPort::~OPGrailPort ()
{
	OP_METHOD (~OPGrailPort, 0);
	
	if (this->info) {
		OPaC_GrailPort_Kill (this->info);
	}
}



/*
 *	Save the graal port to the archive. Is this really
 *	meaningful ?
 */

void
OPGrailPort::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	super::StoreInstance (storage);
	
	//	NIY
}



/*
 *	Restore the graal port from the archive. Is this
 *	really meaningful ?
 */

void
OPGrailPort::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	super::RestoreInstance (storage);
	
	//	NIY
}



/*
 *	Prepare the grail port for deletion.
 */

void
OPGrailPort::Delete ()
{
	OP_METHOD (Delete, 0);
	
	if (this->adorner.IsValid ()) {
	
		//	Just remove the adorner, because it is tightly coupled to the
		//	GrailPort. Be careful not to call Delete for the GrailPort from
		//	the adorner (in fact, it would not harm, just be useless).
		
		OPAdornerRef adorn = this->adorner;
		
		this->adorner = 0;
		adorn->Delete ();
	}
}



/*
 *	Records the application pointer.
 */	

void
OPGrailPort::InitApp (OPAppIn app)
{
	OP_METHOD (InitApp, 1);
	OP_ARG (1, OPAppIn, app);
	this->application = app;
}



/*
 *	Records the adorner. Usually, this will only be called
 *	if an already existing adorner wants to subscribe.
 */

void
OPGrailPort::InitAdorner (OPAdornerIn adorner)
{
	OP_METHOD (InitAdorner, 1);
	OP_ARG (1, OPAdornerIn, adorner);
	this->adorner = adorner;
}



/*
 *	Start drawing into a specified rectangle. The subclass
 *	has to initialise properly the "info->hdc" field so
 *	that the display context has been validated.
 */

void
OPGrailPort::BeginDraw (OPRectIn box, OPRelCoord margin)
{
	OP_METHOD (BeginDraw, 2);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPRelCoord, margin);
	
	if ( ((OPRelCoord)(box.DX ()) < (2*margin))
	  || ((OPRelCoord)(box.DY ()) < (2*margin)) ) {
		return;
	}
	
	OPCoord x1 = box.X1 () + margin;
	OPCoord y1 = box.Y1 () + margin;
	OPCoord x2 = box.X2 () - margin;
	OPCoord y2 = box.Y2 () - margin;
	
	this->BeginDraw (x1, y1, x2, y2);
}

void
OPGrailPort::BeginDraw (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2)
{
	OP_METHOD (BeginDraw, 4);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);
	OP_ASSERT (this->is_ready == TRUE);
	OP_ASSERT (this->is_drawing == FALSE);
	OP_ASSERT ((x2 > x1) && (y2 > y1));
	
	this->rect_x1 = x1;
	this->rect_y1 = y1;
	this->rect_x2 = x2;
	this->rect_y2 = y2;
	
	//	Setup a clipping rectangle so that the user won't inadvertently
	//	draw where (s)he shouldn't.
	
	OPaC_GrailPort_BeginDraw (this->info, (int)(x1), (int)(y1), (int)(x2), (int)(y2),
							  this->color_man->ReturnPrivateInfo ());
	
	//	Restore standard line style, properties and font.
	
	this->is_drawing = TRUE;
	this->SetLineStyle (OP_LINE_SOLID);
	this->SetFont (this->font);
	this->ResetClipRect ();
}



/*
 *	The drawing is finished. The subclass will have to set
 *	the drawn part as dirty, if double buffering.
 */

void
OPGrailPort::EndDraw ()
{
	OP_METHOD (EndDraw, 0);
	OPaC_GrailPort_EndDraw (this->info);
	this->is_drawing = FALSE;
}



/*
 *	Flush the output buffer to synchronize with the screen.
 */

void
OPGrailPort::Flush ()
{
	OP_METHOD (Flush, 0);
	OPaC_GrailPort_Flush (this->info);
}



/*
 *	Define a clipping rectangle, which will truncate every
 *	graphical operation. The new clipping region replaces
 *	the previous one.
 */

void
OPGrailPort::SetClipRect (OPRectIn box, OPRelCoord margin)
{
	OP_METHOD (SetClipRect, 2);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPRelCoord, margin);
	
	if ( ((OPRelCoord)(box.DX ()) < (2*margin))
	  || ((OPRelCoord)(box.DY ()) < (2*margin)) ) {
		OPaC_GrailPort_SetClipRect (this->info, 0, 0, 0, 0);
		return;
	}
	
	OPCoord x1 = box.X1 () + margin;
	OPCoord y1 = box.Y1 () + margin;
	OPCoord x2 = box.X2 () - margin;
	OPCoord y2 = box.Y2 () - margin;
	
	this->SetClipRect (x1, y1, x2, y2);
}

void
OPGrailPort::SetClipRect (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2)
{
	OP_METHOD (SetClipRect, 4);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);
	OP_ASSERT (this->is_drawing);
	
	this->clip_x1 = (x1 > this->rect_x1) ? x1 : this->rect_x1;
	this->clip_y1 = (y1 > this->rect_y1) ? y1 : this->rect_y1;
	this->clip_x2 = (x2 < this->rect_x2) ? x2 : this->rect_x2;
	this->clip_y2 = (y2 < this->rect_y2) ? y2 : this->rect_y2;
	
	OPaC_GrailPort_SetClipRect (this->info, (int)(x1), (int)(y1), (int)(x2), (int)(y2));
}



/*
 *	Return the current clipping rectangle. Default is same
 *	as "BeginDraw" region.
 */

void
OPGrailPort::GetClipRect (OPCoordOut x1, OPCoordOut y1,
						  OPCoordOut x2, OPCoordOut y2) const
{
	OP_METHOD (GetClipRect, 4);
	OP_ARG (1, OPCoordOut, x1);
	OP_ARG (2, OPCoordOut, y1);
	OP_ARG (3, OPCoordOut, x2);
	OP_ARG (4, OPCoordOut, y2);
	
	x1 = this->clip_x1;
	y1 = this->clip_y1;
	x2 = this->clip_x2;
	y2 = this->clip_y2;
}

void
OPGrailPort::GetClipRect (OPRectOut rect) const
{
	OP_METHOD (GetClipRect, 1);
	OP_ARG (1, OPRectOut, rect);
	
	rect = OPRect (this->clip_x1, this->clip_y1, this->clip_x2, this->clip_y2);
}



/*
 *	Reset the clipping rectangle to the default one.
 */

void
OPGrailPort::ResetClipRect ()
{
	OP_METHOD (ResetClipRect, 0);
	
	this->clip_x1 = this->rect_x1;
	this->clip_y1 = this->rect_y1;
	this->clip_x2 = this->rect_x2;
	this->clip_y2 = this->rect_y2;
	
	OPaC_GrailPort_SetClipRect (this->info,
								(int)(this->clip_x1), (int)(this->clip_y1),
								(int)(this->clip_x2), (int)(this->clip_y2));
}


/*
 *	Return a reference to the color manager associated to
 *	this Grail Port.
 */

const OPColorManRef&
OPGrailPort::ReturnColorMan () const
{
	OP_METHOD (ReturnColorMan, 0);
	return this->color_man;
}



/*
 *	Return a pointer to the font manager associated to this
 *	Grail Port.
 */

const OPFontManRef&
OPGrailPort::ReturnFontMan () const
{
	OP_METHOD (ReturnFontMan, 0);
	return this->font_man;
}



/*
 *	Return (and allocate if not already done) the Adorner
 *	associated to the port. This can be used by the widgets
 *	to draw some GUI elements.
 */

const OPAdornerRef&
OPGrailPort::ReturnAdorner ()
{
	OP_METHOD (ReturnAdorner, 0);
	
	if (this->adorner.IsValid () == FALSE) {
		this->adorner = OP_NEW_FROM_ZONE (OPaC_sys_zone, OPAdorner);
		this->adorner->Initialise (this);
	}
	
	return this->adorner;
}



/*
 *	Define the foreground color used when drawing
 *	monochrome objects (lines, circles, text...) and
 *	filling.
 */

void
OPGrailPort::SetForeColor (OPColorPixel pixel)
{
	OP_METHOD (SetForeColor, 1);
	OP_ARG (1, OPColorPixel, pixel);
	
	if (pixel == this->fore_color)
		return;
	
	this->fore_color = pixel;
	this->SetLineStyle (this->line_style);
	
	OPaC_GrailPort_SetForeColor (this->info, pixel, this->color_man);
}



/*
 *	Define the background color.
 */

void
OPGrailPort::SetBackColor (OPColorPixel pixel)
{
	OP_METHOD (SetBackColor, 1);
	OP_ARG (1, OPColorPixel, pixel);
	
	this->back_color = pixel;
	OPaC_GrailPort_SetBackColor (this->info, pixel, this->color_man);
}


/*
 *	Define the transparent color. This is used for the
 *	CopyRect methods (it might be extended later to other
 *	drawing functions).
 */

void
OPGrailPort::SetTransparency (OPColorPixel transparent_color)
{
	OP_METHOD (SetTransparency, 1);
	OP_ARG (1, OPColorPixel, transparent_color);
	
	this->transparent_color = transparent_color;
}


/*
 *	Enable (ok = TRUE) or disable the transparency.
 */

void
OPGrailPort::SetTransparent (Bool ok)
{
	OP_METHOD (SetTransparent, 1);
	OP_ARG (1, Bool, ok);
	
	this->use_transparency = ok;
}


/*
 *	Set up a line style. This will (on Windows) take in account
 *	the foreground color for the line style.
 */

void
OPGrailPort::SetLineStyle (OPLineStyle style)
{
	OP_METHOD (SetLineStyle, 1);
	OP_ARG (1, OPLineStyle, style);
	OP_ASSERT (this->is_drawing);
	
	OPaC_GrailPort_SetLineStyle (this->info, this->fore_color, style, this->color_man);
	
	this->line_style = style;
}



/*
 *	Set up a font. If the specified font pointer is 0, use
 *	a default, fixed-pitch, system font.
 */

void
OPGrailPort::SetFont (OPGrailFontIn font)
{
	OP_METHOD (SetFont, 1);
	OP_ARG (1, OPGrailFontIn, font);
	
/*	if ( (this->font.IsValid ())		*		//	font may be modified by the system, so..
 *	  && (this->font == font) ) {		*		//	..tell low-level about it, even if we..
 *		return;							*		//	..think that it is not needed.
 *	}									*/
	
	//	If the font has been specified, just use it. Otherwise, it will be the
	//	low level interface's responsibility to handle a NULL font.
	
	OPaC_GrFont* font_info = 0;
	
	if (font.IsValid ()) {
		this->font = font;
		font_info  = font->ReturnPrivateInfo ();
	}
	
	//	Select the font...
	
	OPaC_GrailPort_SetFont (this->info, font_info);
}
