/*
 *	icon.cxx
 *
 *	Icon class implementation.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/grail.h>
#include <OPaC/icon.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>
#include <OPaC/pubmacro.h>

/****************************************************************/

OP_IMPLEMENTATION (icon.h);
OP_IMPLEMENT_CLASS (OPIcon);

/****************************************************************/

OP_PUBLISH_MODIFIER (OPIcon, "NormalIcon",  GetNormalIcon,  SetNormalIcon,  "Normal icon",  OPObject, OPAtom);
OP_PUBLISH_MODIFIER (OPIcon, "PressedIcon", GetPressedIcon, SetPressedIcon, "Pressed icon", OPObject, OPAtom);
OP_PUBLISH_MODIFIER (OPIcon, "HilitedIcon", GetHilitedIcon, SetHilitedIcon, "Hilited icon", OPObject, OPAtom);
OP_PUBLISH_MODIFIER (OPIcon, "DimmedIcon",  GetDimmedIcon,  SetDimmedIcon,  "Dimmed icon",  OPObject, OPAtom);

/****************************************************************/


/*
 *	Create an (empty) icon.
 */

OPIcon::OPIcon ()
{
	OP_METHOD (OPIcon, 0);
	
	this->do_press_on_action  = OPLookMan::Current ()->IsPolicyHiliteIcon () == FALSE;
	this->do_hilite_on_action = OPLookMan::Current ()->IsPolicyHiliteIcon () == TRUE;
	
	this->shape &= ~ OP_CTRL_AUTOACTHILITE;
	
	this->icon        = OP_ICON_NONE;
	this->icon_press  = OP_ICON_NONE;
	this->icon_hilite = OP_ICON_NONE;
	this->icon_dim    = OP_ICON_NONE;
	
	this->dx = 0;
	this->dy = 0;
}



/*
 *	Destroy an icon.
 */

OPIcon::~OPIcon ()
{
	OP_METHOD (~OPIcon, 0);
}



/*
 *	Save an icon to the storage.
 */

void
OPIcon::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	//	Save the icon names rather than the icon identifiers. This will allow
	//	a higher flexibility, especially since user defined icons are possible.
	
	OPLookManRef look = OPLookMan::Current ();
	OPAtomRef    name;
	
	look->GetIconName (this->icon, name);			storage << name;
	look->GetIconName (this->icon_press, name);		storage << name;
	look->GetIconName (this->icon_hilite, name);	storage << name;
	look->GetIconName (this->icon_dim, name);		storage << name;
	
	storage << OPStorageSync::mark;
}



/*
 *	Restore an icon from the storage.
 */

void
OPIcon::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	OPLookManRef look = OPLookMan::Current ();
	OPAtomRef    name;
	
	storage >> name; this->icon        = look->FindIcon (name);
	storage >> name; this->icon_press  = look->FindIcon (name);
	storage >> name; this->icon_hilite = look->FindIcon (name);
	storage >> name; this->icon_dim    = look->FindIcon (name);
	
	storage >> OPStorageSync::mark;
	
	OPLookMan::Current ()->GetIconSize (this->icon, this->dx, this->dy);
}



/*
 *	Prepare the icon for its destruction.
 */

void
OPIcon::Delete ()
{
	OP_METHOD (Delete, 0);
	super::Delete ();
}



/*
 *	Initialise an icon.
 */

void
OPIcon::Initialise (OPLookIcon icon)
{
	OP_METHOD (Initialise, 1);
	OP_ARG (1, OPLookIcon, icon);
	
	this->Initialise (icon, icon, icon, icon);
}



/*
 *	Initialise an icon.
 */

void
OPIcon::Initialise (OPLookIcon icon, OPLookIcon press, OPLookIcon hilite, OPLookIcon dim)
{
	OP_METHOD (Initialise, 4);
	OP_ARG (1, OPLookIcon, icon);
	OP_ARG (2, OPLookIcon, press);
	OP_ARG (3, OPLookIcon, hilite);
	OP_ARG (4, OPLookIcon, dim);
	
	this->icon        = icon;
	this->icon_press  = press;
	this->icon_hilite = hilite;
	this->icon_dim    = dim;
	
	OPLookMan::Current ()->GetIconSize (this->icon, this->dx, this->dy);
	
	this->SetSize (this->dx, this->dy);
}


/*
 *	Return the internal icons.
 */

OPLookIcon
OPIcon::ReturnNormalIcon () const
{
	OP_METHOD (ReturnNormalIcon, 0);
	return this->icon;
}

OPLookIcon
OPIcon::ReturnPressedIcon () const
{
	OP_METHOD (ReturnPressedIcon, 0);
	return this->icon_press;
}

OPLookIcon
OPIcon::ReturnHilitedIcon () const
{
	OP_METHOD (ReturnHilitedIcon, 0);
	return this->icon_hilite;
}

OPLookIcon
OPIcon::ReturnDimmedIcon () const
{
	OP_METHOD (ReturnDimmedIcon, 0);
	return this->icon_dim;
}


/*
 *	Set the icons by their names. The mapping between the name
 *	and the internal icon code is done by the look manager.
 */

Bool
OPIcon::SetNormalIcon (OPAtomIn name)
{
	OP_METHOD (SetNormalIcon, 1);
	OP_ARG (1, OPAtomIn, name);
	
	this->icon = OPLookMan::Current ()->FindIcon (name);
	
	return TRUE;
}

Bool
OPIcon::SetPressedIcon (OPAtomIn name)
{
	OP_METHOD (SetPressedIcon, 1);
	OP_ARG (1, OPAtomIn, name);
	
	this->icon_press = OPLookMan::Current ()->FindIcon (name);
	
	return TRUE;
}

Bool
OPIcon::SetHilitedIcon (OPAtomIn name)
{
	OP_METHOD (SetHilitedIcon, 1);
	OP_ARG (1, OPAtomIn, name);
	
	this->icon_hilite = OPLookMan::Current ()->FindIcon (name);
	
	return TRUE;
}

Bool
OPIcon::SetDimmedIcon (OPAtomIn name)
{
	OP_METHOD (SetDimmedIcon, 1);
	OP_ARG (1, OPAtomIn, name);
	
	this->icon_dim = OPLookMan::Current ()->FindIcon (name);
	
	return TRUE;
}


/*
 *	Get the name of the icons... If you want the internal code,
 *	use ReturnXXXIcon instead.
 */

Bool
OPIcon::GetNormalIcon (OPAtomOut name) const
{
	OP_METHOD (GetNormalIcon, 1);
	OP_ARG (1, OPAtomOut, name);
	
	return OPLookMan::Current ()->GetIconName (this->icon, name);
}

Bool
OPIcon::GetPressedIcon (OPAtomOut name) const
{
	OP_METHOD (GetPressedIcon, 1);
	OP_ARG (1, OPAtomOut, name);
	
	return OPLookMan::Current ()->GetIconName (this->icon_press, name);
}

Bool
OPIcon::GetHilitedIcon (OPAtomOut name) const
{
	OP_METHOD (GetHilitedIcon, 1);
	OP_ARG (1, OPAtomOut, name);
	
	return OPLookMan::Current ()->GetIconName (this->icon_hilite, name);
}

Bool
OPIcon::GetDimmedIcon (OPAtomOut name) const
{
	OP_METHOD (GetDimmedIcon, 1);
	OP_ARG (1, OPAtomOut, name);
	
	return OPLookMan::Current ()->GetIconName (this->icon_dim, name);
}



/*
 *	Return the icon's minimal size. This will be the
 *	glyph's size.
 */

void
OPIcon::GetMinSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMinSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	super::GetMinSize (dx, dy);
	
	if (dx < this->dx) dx = this->dx;
	if (dy < this->dy) dy = this->dy;
}


/*
 *	Return the icon's maximum size.
 */

void
OPIcon::GetMaxSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMaxSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	super::GetMaxSize (dx, dy);
}



/*
 *	Draw the icon. This mainly consists in a raster
 *	operation.
 */

void
OPIcon::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	super::Draw (port, rect, what);
	
	if (what & OP_DRAW_OBJECT) {
		
		const OPAdornerRef& adorner = port->ReturnAdorner ();
		
		OPCoord x = this->box.X1 () + (this->box.DX () - this->dx) / 2;
		OPCoord y = this->box.Y1 () + (this->box.DY () - this->dy) / 2;
		
		if ((this->state & OP_VIEW_IS_ENABLED) == 0) {
			adorner->DrawIcon (x, y, this->icon_dim);
		} else if (this->is_being_hilited
				|| this->is_actively_hilited) {
			adorner->DrawIcon (x, y, this->icon_hilite);
		} else if (this->is_being_pressed) {
			adorner->DrawIcon (x, y, this->icon_press);
		} else {
			adorner->DrawIcon (x, y, this->icon);
		}
	}
}



/*
 *	The only message an icon understands is that the look
 *	manager has changed.
 */

void
OPIcon::Broadcast (OPAtomIn message, OPObjectIn object)
{
	OP_METHOD (Broadcast, 2);
	OP_ARG (1, OPAtomIn, message);
	OP_ARG (2, OPObjectIn, object);
	
	if (message == this->atom_look_man_changed) {
		
		this->do_press_on_action  = OPLookMan::Current ()->IsPolicyHiliteButton () == FALSE;
		this->do_hilite_on_action = OPLookMan::Current ()->IsPolicyHiliteButton () == TRUE;
		
		if (this->icon != OP_ICON_NONE) {
			OPLookMan::Current ()->GetIconSize (this->icon, this->dx, this->dy);
			this->SetSize (this->dx, this->dy);
		}
	}
	
	super::Broadcast (message, object);
}

