/*
 *	led7disp.h
 *
 *	The LED 7-segment Display class. This can be used to display
 *	numerical values.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/led7disp.h>

#include <OPaC/grail.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>
#include <OPaC/pubmacro.h>
#include <OPaC/string.h>

/****************************************************************/

OP_IMPLEMENTATION (led7disp.h);
OP_IMPLEMENT_CLASS (OPLed7Display);

/****************************************************************/

#define	MAX_DIGIT_NUM	32

/****************************************************************/

/*
 *	The digits 0..9 and A..F are contiguous. Two more special
 *	digits have been added : the empty space and the minus sign.
 */

#define	DIGIT_EMPTY		16
#define	DIGIT_MINUS		17
#define	DIGIT_NUM		18

/****************************************************************/

OP_PUBLISH_MODIFUNC (OPLed7Display, "Digits", ReturnDigits, SetDigits, "Digits", Card32, Card32);

/****************************************************************/

/*
 *	Create a LED display.
 */

OPLed7Display::OPLed7Display ()
{
	OP_METHOD (OPLed7Display, 0);
	
	this->shape     |= OP_CTRL_RELIEF;
	this->led_color  = OPColorMan::ReturnSystemColorMan ()->FindColor (1.0, 0.0, 0.0);
	this->led_digits = 0;
	this->led_signed = TRUE;
}



/* 
 *	Destroy a LED display.
 */

OPLed7Display::~OPLed7Display ()
{
	OP_METHOD (~OPLed7Display, 0);
}



/*
 *	Store the LED display.
 */

void
OPLed7Display::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->led_color << this->led_digits << this->led_signed;
	storage << OPStorageSync::mark;
}



/*
 *	Restore the LED display.
 */

void
OPLed7Display::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> this->led_color >> this->led_digits >> this->led_signed;
	storage >> OPStorageSync::mark;
}



Bool
OPLed7Display::SetColor (OPColorPixel pixel)
{
	OP_METHOD (SetColor, 1);
	OP_ARG (1, OPColorPixel, pixel);
	
	this->led_color = pixel;
	
	return TRUE;
}

Bool
OPLed7Display::SetDigitsSign (Count digits, Bool use_sign)
{
	OP_METHOD (SetDigitsSign, 2);
	OP_ARG (1, Count, digits);
	OP_ARG (2, Bool, use_sign);
	
	this->led_digits = digits;
	this->led_signed = use_sign;
	
	return TRUE;
}

Bool
OPLed7Display::SetDigits (Count digits)
{
	OP_METHOD (SetDigits, 1);
	OP_ARG (1, Count, digits);
	
	this->led_digits = digits;
	
	return TRUE;
}

Bool
OPLed7Display::SetSign (Bool use_sign)
{
	OP_METHOD (SetSign, 1);
	OP_ARG (1, Bool, use_sign);
	
	this->led_signed = use_sign;
	
	return TRUE;
}



OPColorPixel
OPLed7Display::ReturnColor () const
{
	OP_METHOD (ReturnColor, 0);
	return this->led_color;
}

Count
OPLed7Display::ReturnDigits () const
{
	OP_METHOD (ReturnDigits, 0);
	return this->led_digits;
}

Bool
OPLed7Display::ReturnSigned () const
{
	OP_METHOD (ReturnSigned, 0);
	return this->led_signed;
}



/*
 *	Draw the segments of the LED display which correspond to
 *	the specified digit. Digits 0..9, A..F, ' ' and '-' are
 *	supported.
 */

static void
Draw7Segment (OPGrailPortIn port, OPCoord x, OPCoord y, Byte digit, Bool dot)
{
	const int h0[DIGIT_NUM] = { 1, 0, 1, 1, 0, 1, 1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 0, 0 };
	const int h1[DIGIT_NUM] = { 0, 0, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 1, 0, 1 };
	const int h2[DIGIT_NUM] = { 1, 0, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1, 1, 1, 1, 0, 0, 0 };
	const int lt[DIGIT_NUM] = { 1, 0, 0, 0, 1, 1, 1, 0, 1, 1, 1, 1, 1, 0, 1, 1, 0, 0 };
	const int lb[DIGIT_NUM] = { 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 1, 1, 1, 1, 1, 1, 0, 0 };
	const int rt[DIGIT_NUM] = { 1, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0 };
	const int rb[DIGIT_NUM] = { 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 0, 0, 0, 0 };
	
	if (h0[digit]) port->DrawLineDX (x + 2, y +  0, 5);
	if (h1[digit]) port->DrawLineDX (x + 2, y +  6, 5);
	if (h2[digit]) port->DrawLineDX (x + 2, y + 12, 5);
	
	if (lt[digit]) port->DrawLineDY (x + 1, y +  1, 5);
	if (lb[digit]) port->DrawLineDY (x + 1, y +  7, 5);
	
	if (rt[digit]) port->DrawLineDY (x + 7, y +  1, 5);
	if (rb[digit]) port->DrawLineDY (x + 7, y +  7, 5);
	
	if (dot) port->DrawDot (x + 9, y + 13);
}


/*
 *	Convert the input string into a set of digits, ranging from
 *	0..9, A..F and including the empty space and minus sign.
 *
 *	The digits and dots parameters can be NULL => this will just
 *	count the number of valid digits found.
 */

static void
MakeDigits (OPStringIn text, Byte* digits, Bool* dots, Count& num, Bool force_sign)
{
	LatinChar buffer[MAX_DIGIT_NUM];
	Bool has_sign = FALSE;
	
	for (Count index = 0; index < MAX_DIGIT_NUM; index++) {
		if (digits) digits[index] = DIGIT_EMPTY;
		if (dots)   dots[index]   = FALSE;
	}
	
	num = 0;
	
	if (text.IsValid ()) {
		
		const LatinChar* ptr = buffer;
		text->GetStringCopy (buffer, MAX_DIGIT_NUM);
		buffer[MAX_DIGIT_NUM-1] = 0;
		
		while (*ptr) {
			switch (*ptr) {
				case '0': case '1': case '2': case '3': case '4':
				case '5': case '6': case '7': case '8': case '9':
					if ((num == 0) && (!has_sign) && (force_sign)) num++;
					if (digits) digits[num] = *ptr - '0';
					if (dots) dots[num+1] = FALSE;
					num++;
					break;
				case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
					if ((num == 0) && (!has_sign) && (force_sign)) num++;
					if (digits) digits[num] = *ptr - 'A' + 10;
					if (dots) dots[num+1] = FALSE;
					num++;
					break;
				case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
					if ((num == 0) && (!has_sign) && (force_sign)) num++;
					if (digits) digits[num] = *ptr - 'a' + 10;
					if (dots) dots[num+1] = FALSE;
					num++;
					break;
				case '.': case ',':
					if (dots) dots[num] = TRUE;
					break;
				case '-':
					if (num == 0) {
						has_sign = TRUE;
						if (digits) digits[num] = DIGIT_MINUS;
						if (dots) dots[num+1] = FALSE;
						num++;
					}
					break;
				default:
					break;
			}
			ptr++;
		}
	}
}


/*
 *	Return the LED's minimal size (which is also the
 *	maximum size => exact size).
 */

void
OPLed7Display::GetMinSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMinSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	Count num = 0;
	
	if (this->led_digits) {
		num = this->led_digits;
	} else {
		MakeDigits (this->text, 0, 0, num, this->led_signed);
	}
	
	OPCoord look_frame = OPLookMan::Current ()->Return3DEffectWidth ();
	
	dx = 10 * num + 2*look_frame;
	dy = 12 + 2*look_frame;
}


void
OPLed7Display::GetMaxSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMaxSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	this->GetMinSize (dx, dy);
}


/*
 *	Draw the LED display.
 */

void
OPLed7Display::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	super::Draw (port, rect, what);
	
	if (what & OP_DRAW_OBJECT) {
		
		Byte  digits[MAX_DIGIT_NUM];
		Bool  dots[MAX_DIGIT_NUM];
		Count num = 0;
		
		MakeDigits (this->text, digits, dots, num, this->led_signed);
		
		port->SetForeColor (this->led_color);
		
		while (num--) {
			
			OPCoord x1 = this->box.X1 () + 2 + num*10;
			OPCoord y1 = this->box.Y1 () + 2;
			
			Draw7Segment (port, x1+1, y1+1, digits[num], dots[num+1]);
		}
	}
}
