/*
 *	objbprovider.cxx
 *
 *	The Object Browser Provider helps browsing object data fields.
 *	
 *	(C) Copyright 1997-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/container.h>
#include <OPaC/convert.h>
#include <OPaC/icon.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>
#include <OPaC/objectbrowser.h>
#include <OPaC/string.h>
#include <OPaC/values.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPObjBProvider);

/****************************************************************/


OPObjBProvider::OPObjBProvider ()
{
	OP_METHOD (OPObjBProvider, 0);
	
	this->is_container = FALSE;
}

OPObjBProvider::~OPObjBProvider ()
{
	OP_METHOD (~OPObjBProvider, 0);
}

void
OPObjBProvider::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->object;
	storage << OPStorageSync::mark;
}

void
OPObjBProvider::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> this->object;
	storage >> OPStorageSync::mark;
	
	this->num_fields   = object->CountPublicData ();
	this->is_container = object->IsKindOf (OP_RUNTIME (OPGenericContainer));
}


void
OPObjBProvider::Initialise (OPObjectIn object)
{
	OP_METHOD (Initialise, 1);
	OP_ARG (1, OPObjectIn, object);
	
	this->object       = object;
	this->num_fields   = object->CountPublicData ();
	this->is_container = object->IsKindOf (OP_RUNTIME (OPGenericContainer));
}

Count
OPObjBProvider::ReturnNumObjects ()
{
	OP_METHOD (ReturnNumObjects, 0);
	
	Count total = this->num_fields;
	
	if (this->is_container) {
		OPGenericContainerRef container = this->object;
		total += container->TotalCount ();
	}
	
	return total;
}


Bool
OPObjBProvider::GetObject (Count index, OPObjectOut object)
{
	OP_METHOD (GetObject, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPObjectOut, object);
	
	if (index < this->num_fields) {
		
		if (this->object->QueryData (index, object)) {
			return TRUE;
		}
		
		OPValueRef value;
		
		if (this->object->QueryData (index, value)) {
			object = value;
			return TRUE;
		}
		
		return FALSE;
	}
	
	index -= this->num_fields;
	
	if (this->is_container) {
		OPGenericContainerRef container = this->object;
		return container->Get (index, object);
	}
	
	return FALSE;
}

Bool
OPObjBProvider::GetText (Count index, OPStringOut text)
{
	OP_METHOD (GetText, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPStringOut, text);
	
	OPAtomRef   name;
	OPStringRef num;
	
	if (index < this->num_fields) {
		if (this->object->GetPublicData (index, name)) {
			text = name->ReturnString ();
			return TRUE;
		}
	}
	
	OPConvert::CardToString (index-this->num_fields, OPStringNull, num);
	
	text = "[";
	text->AppendString (num);
	text->AppendString ("]");
	
	return TRUE;
}

Bool
OPObjBProvider::GetLargeIcon (Count index, OPLookIconOut icon)
{
	OP_METHOD (GetLargeIcon, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPLookIconOut, icon);
	
	OPObjectRef object;
	
	if (this->GetObject (index, object)) {
		if (object.IsValid ()) {
			if (object->IsKindOf (OP_RUNTIME (OPApp))) {
				icon = OP_ICON_OBJ_APP;
				return TRUE;
			}
			if (object->IsKindOf (OP_RUNTIME (OPMenu))) {
				icon = OP_ICON_OBJ_MENU;
				return TRUE;
			}
			if (object->IsKindOf (OP_RUNTIME (OPWindow))) {
				icon = OP_ICON_OBJ_WINDOW;
				return TRUE;
			}
			if (object->IsKindOf (OP_RUNTIME (OPDataSlot))) {
				icon = OP_ICON_OBJ_SLOT;
				return TRUE;
			}
			icon = OP_ICON_CUBE_NORMAL;
			return TRUE;
		}
	}
	
	return FALSE;
}

Bool
OPObjBProvider::GetSmallIcon (Count index, OPLookIconOut icon)
{
	OP_METHOD (GetSmallIcon, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPLookIconOut, icon);
	
	return FALSE;
}

/*
 *	Return TRUE if the object might have other embedded
 *	objects. The "real" object will be returned if the
 *	original object was a value.
 */

Bool
OPObjBProvider::FindDeepObject (OPObjectInOut object)
{
	OP_METHOD (FindDeepObject, 1);
	OP_ARG (1, OPObjectInOut, object);
	
	if (object.IsValid ()) {
		
		//	Handle the special case of the value object... which is either a value
		//	by itself or embeds another (possibly complex) object.
		
		while (object->IsKindOf (OP_RUNTIME (OPValue))) {
			
			OPValueRef value = object;
			
			switch (value->ReturnType ()) {
				case OP_TYPE_OBJECT:
				case OP_TYPE_TEXT:
				case OP_TYPE_ARRAY:
					break;
				default:
					return FALSE;
			}
			
			object = value->ReturnObject ();
		}
		
		if (object->IsKindOf (OP_RUNTIME (OPString))) {
			return FALSE;
		}
		
		return TRUE;
	}
	
	return FALSE;
}

Bool
OPObjBProvider::GetMarkIcon (Count index, OPLookIconOut icon)
{
	OP_METHOD (GetMarkIcon, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPLookIconOut, icon);
	
	OPObjectRef object;
	
	if ( (this->GetObject (index, object))
	  && (object.IsValid ())
	  && (object->CountPublicData ())
	  && (this->FindDeepObject (object)) ) {
		
		icon = OP_ICON_SUBMENU_NORMAL;
		
		return TRUE;
	}
	
	return FALSE;
}


/*
 *	Build the subtree provider for the specified entry. Return
 *	FALSE if this is not possible.
 */

Bool
OPObjBProvider::EnterSubtree (Count index, OPGenericProviderOut provider)
{
	OP_METHOD (EnterSubtree, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPGenericProviderOut, provider);
	
	OPObjectRef object;
	
	if ( (this->GetObject (index, object))
	  && (object.IsValid ())
	  && (object->CountPublicData ())
	  && (this->FindDeepObject (object)) ) {
		
		OPObjBProviderRef pro = OP_NEW (OPObjBProvider);
		
		pro->Initialise (object);
		provider = pro;
		
		return TRUE;
	}
	
	return FALSE;
}

Bool
OPObjBProvider::ReleaseSubtree (Count index)
{
	OP_METHOD (ReleaseSubtree, 1);
	OP_ARG (1, Count, index);
	
	return TRUE;
}

