/*
 *	rtclass.cxx
 *
 *	This file implements the run-time OPaC information class. This
 *	is thightly coupled to the OPObject and OPClassMan classes.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/except.h>
#include <OPaC/new.h>
#include <OPaC/object.h>
#include <OPaC/opacirt.h>
#include <OPaC/notmacro.h>

#include <string.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPRuntime);

/****************************************************************/

OPaC_PubDataExport* OPaC_PubDataExport::head = 0;
OPaC_PubDataExport* OPaC_PubDataExport::tail = 0;

OPaC_NotifyExport*	OPaC_NotifyExport::head  = 0;
OPaC_NotifyExport*	OPaC_NotifyExport::tail	 = 0;

/****************************************************************/

/*
 *	Notification events can be registered in the run-time
 *	class, so that a caller can find out dynamically which
 *	events an observer might be interested in.
 */

struct OPRuntime::NotifyInfo
{
	NotifyInfo*		next;					//	next notification event info
	OPAtomRef		name;					//	name of notification event
	OPAtomRef		field;					//	associated field (if any)
};

/****************************************************************/

/*
 *	Once the start-up has succeeded, OPClassMan::Init will
 *	instanciate the run-time descriptions for each class.
 */
	
OPRuntime::OPRuntime ()
{
	OP_METHOD (OPRuntime, 0);
	
	this->self        = this;
	this->super_class = 0;
	this->info        = 0;
	
	this->target_list = 0;
	this->target_num  = 0;
	this->data_list   = 0;
	this->data_num    = 0;
	this->data_array  = 0;
	this->notify_list = 0;
	this->notify_num  = 0;
	
	this->target_cache.target  = 0;
	this->target_cache.index   = 0;
	this->target_cache.runtime = 0;
}



/*
 *	We still don't do anything special on run-time info
 *	destruction.
 */

OPRuntime::~OPRuntime ()
{
	OP_METHOD (~OPRuntime, 0);
}



/*
 *	Prepare the run-time class for deletion.
 */

void
OPRuntime::Delete ()
{
	OP_METHOD (Delete, 0);
	
	//	Remove the targets recorded for this run-time class :
	
	if (this->target_list) {
		OPaC_Target* iter = this->target_list;
		OPaC_Target* next = 0;
		
		while (iter) {
			next = iter->next;
			DELETE (iter);
			iter = next;
		}
	}
	
	//	Remove the published data recorded for this run-time class :
	
	if (this->data_list) {
		OPaC_PubData* iter = this->data_list;
		OPaC_PubData* next = 0;
		OPaC_PubData* simi = 0;
		
		while (iter) {
			
			//	If there are similar data definitions, they are linked in a
			//	special list -- the "sim"-list. First free that one.
			
			simi = iter->sim_next;
			while (simi) {
				next = simi->sim_next;
				DELETE (simi);
				simi = next;
			}
			
			next = iter->next;
			DELETE (iter);
			iter = next;
		}
	}
	
	if (this->data_array) {
		OP_FREE_FROM_ZONE (OPaC_sys_zone, this->data_array);
	}
	
	this->target_list = 0;
	this->target_num  = 0;
	this->data_list   = 0;
	this->data_num    = 0;
	this->data_array  = 0;
	
	//	Remove reference to super-class and then reference to self...
	
	this->super_class   = 0;
	this->self          = 0;
	this->next          = 0;
	
	//	Finally remove the hard-wired reference which had been set up by the
	//	initialization method...
	
	(info->runtime)[0]  = 0;
	this->info->runtime = 0;
}



/*
 *	After creation, the run-time class is initialised.
 */

void
OPRuntime::Initialise (OPaC_RuntimeInfo* info)
{
	OP_METHOD (Initialise, 1);
	OP_ARG (1, const OPaC_RuntimeInfo*, info);
	
	this->info        = info;
	
	//	Now that the run-time information is ready to use, we can initialise the
	//	reference in the body of the class (obtained through OP_RUNTIME).
	
	(info->runtime)[0] = this;
}



/*
 *	After initialization, the run-time class gets linked
 *	to the super-class.
 */

void
OPRuntime::LinkSuper ()
{
	OP_METHOD (LinkSuper, 0);
	
	if (this->info->runtime_func) {
		this->super_class = (this->info->runtime_func)();
	} else {
		this->super_class = 0;
	}
}



/*
 *	Collect data about the targets.
 */

void
OPRuntime::CollectTargets ()
{
	OP_METHOD (CollectTargets, 0);
	
	OPaC_TargetExport*  iter  = OPaC_TargetExport::head;
	OPaC_PubDataExport*	pubd  = OPaC_PubDataExport::head;
	OPaC_NotifyExport*	notif = OPaC_NotifyExport::head;
	
	while (iter) {
		
		const OPaC_TargetMethod* method = &iter->info;
		
		//	Method information has been filled. A pointer to it is being
		//	returned to the caller.
		
		if (method->runtime[0] == this->self) {
			
			//	The target which has been scanned belongs to our run-time
			//	class. We now need to allocate a real target description
			//	for it !
			
			OPaC_Target* info = NEW_FROM_ZONE (OPaC_sys_zone)
								OPaC_Target (this->self, method);
			
			info->next = this->target_list;
			info->Initialise ();
			
			this->target_list = info;
			this->target_num++;
		}
		
		iter = iter->next;
	}

	//	Look for the published data.
	
	while (pubd) {
		(pubd->init)(this);
		pubd = pubd->next;
	}

	//	Look for the registered notifications.

	while (notif) {
		
		if (notif->runtime[0] == this->self) {
			
			NotifyInfo* info = NEW_FROM_ZONE (OPaC_sys_zone) NotifyInfo;
			
			info->next  = this->notify_list;
			info->name  = notif->name;
			info->field = notif->field;
			
			if (notif->field) {
				notif->atom[0] = info->field;
			} else {
				notif->atom[0] = info->name;
			}
			
			this->notify_list = info;
			this->notify_num++;
		}
		
		notif = notif->next;
	}
}


/*
 *	This method gets called for each class which exports
 *	published data.
 */

void
OPRuntime::PublishData (OPaC_PubDataDef& def)
{
	OP_METHOD (PublishData, 1);
	OP_ARG (1, OPaC_PubDataDef&, def);

	//	This is strange... Why are we not called with the proper runtime
	//	class ?
	
	if (def.runtime[0] != this) {
		def.runtime[0]->PublishData (def);
		return;
	}
	
	OPaC_PubData* info = NEW_FROM_ZONE (OPaC_sys_zone) OPaC_PubData (def);
	
	info->Initialise ();
	
	//	Is there another published data definition with the
	//	same name and data type ?
	
	OPaC_PubData* iter = this->data_list;

	while (iter) {
		
		if ( (iter->name == info->name)
		  && (iter->data_def.var_type == info->data_def.var_type) ) {
			
			//	Found similar data definition. Just link with that one
			//	in the list. This won't be counted, since it is just
			//	another way to access the same data.
			
			info->next = 0;
			info->sim_next = iter->sim_next;
			iter->sim_next = info;

			return;
		}
		
		iter = iter->next;
	}

	//	Not found similar data definition, just insert into list. This
	//	represents one more published data...
	
	info->next = this->data_list;
	
	this->data_list = info;
	this->data_num++;
}


void
OPRuntime::PublishDataEx (OPaC_PubDataDef& pub, const OPaC_BitsetDataDef* def, Card32 size)
{
	OP_METHOD (PublishDataEx, 3);
	OP_ARG (1, OPaC_PubDataDef&, pub);
	OP_ARG (2, const OPaC_BitsetDataDef*, def);
	OP_ARG (3, Card32, size);
	
	//	This is strange... Why are we not called with the proper runtime
	//	class ?
	
	if (pub.runtime[0] != this) {
		pub.runtime[0]->PublishDataEx (pub, def, size);
		return;
	}
	
	OPaC_PubData* iter = this->data_list;
	
	while (iter && (strcmp (iter->data_def.var_name, pub.var_name) != 0)) {
		iter = iter->next;
	}
	
	OP_ASSERT (iter);
	
	OPaC_PubData::BitsetData* data = iter->bitset_list;

	while (data) {
		if (data->def == def) {
			return;
		}
		data = data->next;
	}
	
	data = NEW_FROM_ZONE (OPaC_sys_zone) OPaC_PubData::BitsetData;
	
	data->next = iter->bitset_list;
	data->def  = def;
	data->num  = size / sizeof (OPaC_BitsetDataDef);
	
	iter->bitset_list = data;
	iter->bitset_num++;
}


/*
 *	Construct an object and return TRUE if it succeeded.
 *	This might raise an exception.
 */

Bool
OPRuntime::ConstructObject (void* ptr, Card32 size) const
{
	OP_METHOD (ConstructObject, 2);
	OP_ARG (1, void*, ptr);
	OP_ARG (2, Card32, size);
	
	OP_ASSERT (ptr != 0);
	OP_ASSERT (size == this->info->instance_size);
	
	(this->info->constr_func)(ptr, size);
	
	return TRUE;
}



/*
 *	Return the nth target of the run-time class. This does
 *	a sequential search, but it maintains a cache to speed
 *	up the progressive searchs.
 */

Bool
OPRuntime::GetTarget (Count index, OPAtomOut name, OPTargetTypeOut the_type,
					  const OPaC_Target* & the_target)
{
	OP_METHOD (GetTarget, 4);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPAtomOut, name);
	OP_ARG (3, OPTargetTypeOut, the_type);
	OP_ARG (4, const OPaC_Target*&, the_target);
	
	if (index >= this->target_num) {
		return FALSE;
	}
	
	OPaC_Target* iter = this->target_list;
	Count        num  = index;
	
	//	By using a cache, we are able to reduce the access time, since we have
	//	some known starting point. The cache exists for every run-time class.
	
	if ( (this->target_cache.index   <= index)
	  && (this->target_cache.target  != 0)
	  && (this->target_cache.runtime == this) ) {
		
		iter = this->target_cache.target;
		num -= this->target_cache.index;
	}
	
	while (num--) {
		OP_ASSERT (iter->next);
		iter = iter->next;
	}
	
	//	Update the cache, so that a linear increasing search will be sped up.
	//	Return the target's name.
	
	this->target_cache.target  = iter;
	this->target_cache.index   = index;
	this->target_cache.runtime = this;
	
	name = iter->name;

	the_target = iter;
	the_type   = (iter->target_method->m_trampo ? OP_TARGET_IS_METHOD
			   : (iter->target_method->o_trampo ? OP_TARGET_IS_OBSERVER
			   : (iter->target_method->a_trampo ? OP_TARGET_IS_ACTOR : OP_TARGET_IS_INVALID)));

	return TRUE;
}



/*
 *	Return the nth published data of the run-time class.
 *	This does a sequential search, but it maintains a cache
 *	to speed up the progressive searchs.
 */

Bool
OPRuntime::GetPubData (Count index, const OPaC_PubData* & data)
{
	OP_METHOD (GetPubData, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, const OPaC_PubData* &, data);
	
	if (index >= this->data_num) {
		return FALSE;
	}
	
	if (this->data_array == 0) {
		
		OPaC_PubData* iter = this->data_list;
		const Count   num  = this->data_num;
		Count         idx  = 0;
		
		this->data_array = OP_ALLOC_ARRAY_FROM_ZONE (OPaC_sys_zone, OPaC_PubData*, num);
		
		while (idx < num) {
			this->data_array[idx] = iter;
			iter = iter->next;
			idx += 1;
		}
	}
	
	data = this->data_array[index];
	return TRUE;
}


/*
 *	Return the nth exported notification event of the run-time
 *	class. This does a sequential search.
 */

Bool
OPRuntime::GetNotify (Count index, OPAtomOut name)
{
	OP_METHOD (GetNotify, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPAtomOut, name);

	if (index >= this->notify_num) {
		return FALSE;
	}

	NotifyInfo* info = this->notify_list;
	
	while ( (index--) && (info) ) {
		info = info->next;
	}
	
	if (info) {
		name = info->name;
		return TRUE;
	}

	return FALSE;
}

Bool
OPRuntime::GetNotify (OPAtomIn field, Count index, OPAtomOut name)
{
	OP_METHOD (GetNotify, 3);
	OP_ARG (1, OPAtomIn, field);
	OP_ARG (2, Count, index);
	OP_ARG (3, OPAtomOut, name);

	Count       num  = 0;
	NotifyInfo* info = this->notify_list;
	
	while (info) {
		if (info->field == field) {
			if (index-- == 0) {
				name = info->name;
				return TRUE;
			}
			num++;
		}
		info = info->next;
	}
	
	return FALSE;
}


/*
 *	Count how many targets correspond to the specified
 *	name... Don't take in account the observers and the
 *	actors.
 */

Count
OPRuntime::CountMethods (OPAtomIn name) const
{
	OP_METHOD (CountTargets, 1);
	OP_ARG (1, OPAtomIn, name);
	
	if (name.IsInvalid ()) {
		return 0;
	}
	
	Count            num  = 0;
	const OPRuntime* that = this;
	
	//	The search begins with the current class. It will start anew with the
	//	super-class, until we reach the top of the object hierarchy.
	
again:
	OPaC_Target* target = that->target_list;
	
	while (target) {
		if ( (target->target_method->m_trampo)
		  && (target->name == name) ) {
			num++;
		}
		target = target->next;
	}
	
	if ( (target == 0)
	  && (that->super_class.IsValid ()) ) {
		
		//	Search the super-class. We could have done this in a recursive fashion
		//	but it is sure to be more efficient using an ugly "goto" !
		
		that = that->super_class.ReturnPtr ();
		goto again;
	}
	
	return num;
}

/*
 *	Count how many notifiers correspond to the specified
 *	field.
 */

Count
OPRuntime::CountNotify (OPAtomIn field) const
{
	OP_METHOD (CountNotify, 1);
	OP_ARG (1, OPAtomIn, field);
	
	if (field.IsInvalid ()) {
		return 0;
	}
	
	Count num = 0;
	NotifyInfo* info = this->notify_list;
	
	while (info) {
		if (info->field == field) {
			num++;
		}
		info = info->next;
	}
	
	return num;
}



/*
 *	Find the target description given by its name. There
 *	can be more than one, since the parameters might be
 *	quite different.
 *
 *	Don't take in account the observers and the actors.
 */

const OPaC_Target*
OPRuntime::FindMethod (OPAtomIn name, Count nth) const
{
	OP_METHOD (FindMethod, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, int, nth);

	if (name.IsValid () == FALSE) {
		return 0;
	}
	
	const OPRuntime* that = this;
	
	//	The search begins with the current class. If it fails, it will start anew
	//	with the super-class, until we reach the top of the object hierarchy.
	
again:
	OPaC_Target* target = that->target_list;
	
	while (target) {
		if ( (target->target_method->m_trampo)
		  && (target->name == name)
		  && (nth-- == 0) ) {
			break;
		}
		
		target = target->next;
	}
	
	if ( (target == 0)
	  && (that->super_class.IsValid ()) ) {
		
		//	Search the super-class. We could have done this in a recursive fashion
		//	but it is sure to be more efficient using an ugly "goto" !
		
		that = that->super_class.ReturnPtr ();
		goto again;
	}
	
	return target;
}



/*
 *	Find the target description given by the observer's
 *	name. Only one can exist...
 */

const OPaC_Target*
OPRuntime::FindObserver (OPAtomIn name) const
{
	OP_METHOD (FindObserver, 1);
	OP_ARG (1, OPAtomIn, name);

	if (name.IsValid () == FALSE) {
		return 0;
	}
	
	const OPRuntime* that = this;
	
	//	The search begins with the current class. If it fails, it will start anew
	//	with the super-class, until we reach the top of the object hierarchy.
	
again:
	OPaC_Target* target = that->target_list;
	
	while (target) {
		if ( (target->target_method->o_trampo)
		  && (target->name == name) ) {
			break;
		}

		target = target->next;
	}
	
	if ( (target == 0)
	  && (that->super_class.IsValid ()) ) {
		
		//	Search the super-class. We could have done this in a recursive fashion
		//	but it is sure to be more efficient using an ugly "goto" !
		
		that = that->super_class.ReturnPtr ();
		goto again;
	}
	
	return target;
}



/*
 *	Find the target description given by the actor's
 *	name. Only one can exist...
 */

const OPaC_Target*
OPRuntime::FindActor (OPAtomIn name) const
{
	OP_METHOD (FindActor, 1);
	OP_ARG (1, OPAtomIn, name);

	if (name.IsValid () == FALSE) {
		return 0;
	}
	
	const OPRuntime* that = this;
	
	//	The search begins with the current class. If it fails, it will start anew
	//	with the super-class, until we reach the top of the object hierarchy.
	
again:
	OPaC_Target* target = that->target_list;
	
	while (target) {
		if ( (target->target_method->a_trampo)
		  && (target->name == name) ) {
			break;
		}

		target = target->next;
	}
	
	if ( (target == 0)
	  && (that->super_class.IsValid ()) ) {
		
		//	Search the super-class. We could have done this in a recursive fashion
		//	but it is sure to be more efficient using an ugly "goto" !
		
		that = that->super_class.ReturnPtr ();
		goto again;
	}
	
	return target;
}



/*
 *	Find the published data description given by its name.
 *	There can be more than one way to access the published
 *	data, so it may be necessary to call this method more
 *	than once.
 */

const OPaC_PubData*
OPRuntime::FindPubData (OPAtomIn name, Count nth) const
{
	OP_METHOD (FindPubData, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, int, nth);

	if (name.IsValid () == FALSE) {
		return 0;
	}
	
	const OPRuntime* that = this;
	
	//	The search begins with the current class. If it fails, it will start anew
	//	with the super-class, until we reach the top of the object hierarchy.
	
again:
	OPaC_PubData* data = that->data_list;
	
	while (data) {
		if ( (data->name == name)
		  && (nth-- == 0) ) {
			break;
		}
		
		data = data->next;
	}
	
	if ( (data == 0)
	  && (that->super_class.IsValid ()) ) {
		
		//	Search the super-class. We could have done this in a recursive fashion
		//	but it is sure to be more efficient using an ugly "goto" !
		
		that = that->super_class.ReturnPtr ();
		goto again;
	}
	
	return data;
}



/*
 *	Look if this class belongs to the specified class.
 */

Bool
OPRuntime::IsKindOf (OPRuntimeIn what) const
{
	OP_METHOD (IsKindOf, 1);
	OP_ARG (1, OPRuntimeIn, what);
	
	OPRuntimeRef runtime = const_cast<OPRuntime*> (this);
	
	while (runtime.IsValid ()) {
		
		if (runtime->ReturnClassName () == what->ReturnClassName ()) {
			return TRUE;
		}
		
		runtime = runtime->ReturnSuper ();
	}
	
	return FALSE;
}



/*
 *	Return TRUE if the runtime class corresponds to the
 *	specified class.
 */

Bool
OPRuntime::IsA (OPRuntimeIn what) const
{
	OP_METHOD (IsA, 1);
	OP_ARG (1, OPRuntimeIn, what);
	
	if (this->ReturnClassName () == what->ReturnClassName ()) {
		return TRUE;
	}
	
	return FALSE;
}

