/*
 *	selgrid.cxx
 *
 *	The Selection Grid class allows a user to select one or several cells
 *	in a Cell Grid.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/boxview.h>
#include <OPaC/carbon.h>
#include <OPaC/control.h>
#include <OPaC/dragview.h>
#include <OPaC/event.h>
#include <OPaC/grail.h>
#include <OPaC/lookman.h>
#include <OPaC/message.h>
#include <OPaC/new.h>
#include <OPaC/notmacro.h>
#include <OPaC/selgrid.h>
#include <OPaC/targetmacro.h>

/****************************************************************/

OP_IMPLEMENTATION (selgrid.h);
OP_IMPLEMENT_CLASS (OPSelGrid);
	
/****************************************************************/

OP_NOTIFY_EVENT (OPSelGrid, CellActivated);
OP_NOTIFY_EVENT (OPSelGrid, CellDeactivated);
OP_NOTIFY_EVENT (OPSelGrid, CellDragged);

OP_TARGET_OBSERVER (OPSelGrid, HandleCellClick);
OP_TARGET_OBSERVER (OPSelGrid, HandleDragBegin);

OP_TARGET_ACTOR (OPSelGrid, HandleDragCancel);
OP_TARGET_ACTOR (OPSelGrid, HandleDragCommit);

/****************************************************************/

#define	OVERLAY_NONE	0
#define	OVERLAY_BOX		1
#define	OVERLAY_LEFT	2
#define	OVERLAY_RIGHT	3
#define	OVERLAY_ABOVE	4
#define	OVERLAY_BELOW	5

/****************************************************************/

/*
 *	Initialise an empty selection grid.
 */

OPSelGrid::OPSelGrid ()
{
	OP_METHOD (OPSelGrid, 0);
	
	this->state |= OP_VIEW_OK_DETECT;
	
	this->select_num = 0;
	
	this->select_on_click     = FALSE;
	this->select_multiple_ok  = FALSE;
	this->select_at_least_one = FALSE;
	this->search_by_lines     = TRUE;
	
	this->enable_cell_drag = FALSE;
	
	this->is_dragging   = FALSE;
	this->is_drop_valid = FALSE;
	this->is_drop_done  = FALSE;
	this->is_drag_done  = FALSE;
	
	this->insert_between_lines   = TRUE;
	this->insert_between_columns = TRUE;
	this->insert_over_cell       = TRUE;
	
	this->h_percent = 0;
	this->v_percent = 0;
	
	this->draw_overlay = OVERLAY_NONE;
}


/*
 *	Free a selection grid.
 */

OPSelGrid::~OPSelGrid ()
{
	OP_METHOD (~OPSelGrid, 0);
}


/*
 *
 */

void
OPSelGrid::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->select_num;
	storage << this->select_on_click;
	storage << this->select_multiple_ok;
	storage << this->select_at_least_one;
	storage << this->search_by_lines;
	
	storage << this->insert_between_lines;
	storage << this->insert_between_columns;
	storage << this->insert_over_cell;
	
	storage << OPStorageSync::mark;
}



void
OPSelGrid::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	this->ResetDragDropStatus ();
	
	super::RestoreInstance (storage);

	storage >> this->select_num;
	storage >> this->select_on_click;
	storage >> this->select_multiple_ok;
	storage >> this->select_at_least_one;
	storage >> this->search_by_lines;
	
	storage >> this->insert_between_lines;
	storage >> this->insert_between_columns;
	storage >> this->insert_over_cell;
	
	storage >> OPStorageSync::mark;
}


void
OPSelGrid::NotifyArraySize (Count dx, Count dy)
{
	OP_METHOD (NotifyArraySize, 2);
	OP_ARG (1, Count, dx);
	OP_ARG (2, Count, dy);
	
	super::NotifyArraySize (dx, dy);
}

void
OPSelGrid::NotifyAttachCell (OPControlIn cell, Count x, Count y)
{
	OP_METHOD (NotifyAttachCell, 3);
	OP_ARG (1, OPControlIn, cell);
	OP_ARG (2, Count, x);
	OP_ARG (3, Count, y);
	
	super::NotifyAttachCell (cell, x, y);
	
	Count pos = x * this->v_cell_count + y;
	
	if (cell.IsValid ()) {
		
		static OPAtomRef event_OnClick = "OnClick";
		static OPAtomRef event_OnPress = "LeftButtonPress";
		static OPAtomRef actor_HandleCellClick = "HandleCellClick";
		static OPAtomRef actor_HandleDragBegin = "HandleDragBegin";
		
		cell->AttachObserver (this, event_OnClick, actor_HandleCellClick, pos);
		
		if (this->enable_cell_drag) {
			cell->AttachObserver (this, event_OnPress, actor_HandleDragBegin, pos);
		}
	}
}

void
OPSelGrid::NotifyDetachCell (OPControlIn cell, Count x, Count y)
{
	OP_METHOD (NotifyDetachCell, 3);
	OP_ARG (1, OPControlIn, cell);
	OP_ARG (2, Count, x);
	OP_ARG (3, Count, y);
	
	super::NotifyDetachCell (cell, x, y);
	
	Count pos = x * this->v_cell_count + y;
	
	if (cell.IsValid ()) {
		
		static OPAtomRef event_OnClick = "OnClick";
		static OPAtomRef event_OnPress = "LeftButtonPress";
		static OPAtomRef actor_HandleCellClick   = "HandleCellClick";
		static OPAtomRef actor_HandleDragBegin = "HandleDragBegin";
		
		cell->DetachObserver (this, event_OnClick, actor_HandleCellClick, pos);
		
		if (this->enable_cell_drag) {
			cell->DetachObserver (this, event_OnPress, actor_HandleDragBegin, pos);
		}
	}
}


/*
 *	Change the hilite state of a cell. If only one cell can
 *	be active at a given time, handle the case. Also handle
 *	the "at least one" policy.
 */

Bool
OPSelGrid::SetActiveHilite (Count x, Count y, Bool active, Bool report)
{
	OP_METHOD (SetActiveHilite, 4);
	OP_ARG (1, Count, x);
	OP_ARG (2, Count, y);
	OP_ARG (3, Bool, active);
	OP_ARG (4, Bool, report);
	
	if (active) {
		
		if ( (this->select_multiple_ok == TRUE)
		  || (this->select_num == 0) ) {
			
			return this->SetInternalActiveHilite (x, y, TRUE, report);
			
		} else if (this->select_multiple_ok == FALSE) {
			
			Count old_x, old_y;
			
			if (this->FindActiveHilite (old_x, old_y)) {
				this->SetInternalActiveHilite (old_x, old_y, FALSE, report);
			}
			
			return this->SetInternalActiveHilite (x, y, TRUE, report);
		}
		
	} else {
		
		if ( (this->select_at_least_one == FALSE)
		  || (this->select_num > 1) ) {
			
			return this->SetInternalActiveHilite (x, y, FALSE, report);
		}
	}
	
	return FALSE;
}



/*
 *	Silently remove every hilite...
 */

Bool
OPSelGrid::RemoveEveryHilite ()
{
	OP_METHOD (RemoveEveryHilite, 0);
	
	Count x, y;
	
	while ( (this->FindActiveHilite (x, y))
		 && (this->SetInternalActiveHilite (x, y, FALSE, FALSE)) ) {
	}
	
	return TRUE;
}


/*
 *	Change the hilite state of a cell. Don't take care of
 *	the restrictions (only one, at least one, etc.)
 */

Bool
OPSelGrid::SetInternalActiveHilite (Count x, Count y, Bool yes, Bool report)
{
	OP_METHOD (SetInternalActiveHilite, 4);
	OP_ARG (1, Count, x);
	OP_ARG (2, Count, y);
	OP_ARG (3, Bool, yes);
	OP_ARG (4, Bool, report);
	
	if ( (x < this->h_cell_count)
	  && (y < this->v_cell_count) ) {
		
		OPControlRef cell;
		
		if (this->GetCell (x, y, cell)) {
			
			if (cell->ReturnActiveHilite () == yes) {
				return TRUE;
			}
			
			cell->SetActiveHilite (yes);
			this->select_num += (yes) ? 1 : -1;
			
			if (report) {
				this->Report (yes ? notify_CellActivated : notify_CellDeactivated);
			}
			
			cell->Display ();
		}
		
		return TRUE;
	}
	
	return FALSE;
}


/*
 *	Return TRUE if the specified cell is hilited.
 */

Bool
OPSelGrid::ReturnActiveHilite (Count x, Count y) const
{
	OP_METHOD (ReturnActiveHilite, 3);
	OP_ARG (1, Count, x);
	OP_ARG (2, Count, y);
	
	OPControlRef cell;
	
	if ( (x < this->h_cell_count)
	  && (y < this->v_cell_count)
	  && (this->GetCell (x, y, cell)) ) {
		return cell->ReturnActiveHilite ();
	}
	
	return FALSE;
}


/*
 *	Attach or detach a cell observer for each cell in
 *	the grid.
 */

void
OPSelGrid::ChangeCellObserver (OPAtomIn event, OPAtomIn actor, Bool set)
{
	OP_METHOD (ChangeCellObserver, 3);
	OP_ARG (1, OPAtomIn, event);
	OP_ARG (2, OPAtomIn, actor);
	OP_ARG (3, Bool, set);
	
	OPControlRef cell;
	Count pos = 0;
	
	for (Count x = 0; x < this->h_cell_count; x++) {
		for (Count y = 0; y < this->v_cell_count; y++) {
			if (this->GetCell (x, y, cell)) {
				if (set) {
					cell->AttachObserver (this, event, actor, pos);
				} else {
					cell->DetachObserver (this, event, actor, pos);
				}
			}
			pos++;
		}
	}
}


/*
 *	Enable or disable the dragging of individual cells.
 *	This registers an observer which notifies the grid
 *	users about a drag operation...
 */

Bool
OPSelGrid::EnableDragging (Bool yes)
{
	OP_METHOD (EnableDragging, 1);
	OP_ARG (1, Bool, yes);
	
	if (this->enable_cell_drag != yes) {
		this->ChangeCellObserver ("LeftButtonPress", "HandleDragBegin", yes);
		this->enable_cell_drag = yes;
	}
	
	return TRUE;
}

Bool
OPSelGrid::ReturnDraggingEnabled () const
{
	OP_METHOD (ReturnDraggingEnabled, 0);
	return this->enable_cell_drag;
}

Bool
OPSelGrid::IsDragging () const
{
	OP_METHOD (IsDragging, 0);
	return this->is_dragging;
}

Bool
OPSelGrid::SetSelectOnClick (Bool yes)
{
	OP_METHOD (SetSelectOnClick, 1);
	OP_ARG (1, Bool, yes);
	this->select_on_click = yes;
	return TRUE;
}

Bool
OPSelGrid::SetSelectMultipleOk (Bool yes)
{
	OP_METHOD (SetSelectMultipleOk, 1);
	OP_ARG (1, Bool, yes);
	this->select_multiple_ok = yes;
	return TRUE;
}

Bool
OPSelGrid::SetSelectAtLeastOne (Bool yes)
{
	OP_METHOD (SetSelectAtLeastOne, 1);
	OP_ARG (1, Bool, yes);
	this->select_at_least_one = yes;
	return TRUE;
}

Bool
OPSelGrid::SetSearchLinesFirst (Bool yes)
{
	OP_METHOD (SetSearchLinesFirst, 1);
	OP_ARG (1, Bool, yes);
	this->search_by_lines = yes;
	return TRUE;
}

Bool
OPSelGrid::SetInsertBetweenLines (Bool yes)
{
	OP_METHOD (SetInsertBetweenLines, 1);
	OP_ARG (1, Bool, yes);
	this->insert_between_lines = yes;
	return TRUE;
}

Bool
OPSelGrid::SetInsertBetweenColumns (Bool yes)
{
	OP_METHOD (SetInsertBetweenColumns, 1);
	OP_ARG (1, Bool, yes);
	this->insert_between_columns = yes;
	return TRUE;
}

Bool
OPSelGrid::SetInsertOverCell (Bool yes)
{
	OP_METHOD (SetInsertOverCell, 1);
	OP_ARG (1, Bool, yes);
	this->insert_over_cell = yes;
	return TRUE;
}

Bool
OPSelGrid::ReturnSelectOnClick () const
{
	OP_METHOD (ReturnSelectOnClick, 0);
	return this->select_on_click;
}

Bool
OPSelGrid::ReturnSelectMultipleOk () const
{
	OP_METHOD (ReturnSelectMultipleOk, 0);
	return this->select_multiple_ok;
}

Bool
OPSelGrid::ReturnSelectAtLeastOne () const
{
	OP_METHOD (ReturnSelectAtLeastOne, 0);
	return this->select_at_least_one;
}

Bool
OPSelGrid::ReturnSearchLinesFirst () const
{
	OP_METHOD (ReturnSearchLinesFirst, 0);
	return this->search_by_lines;
}

Bool
OPSelGrid::ReturnInsertBetweenLines () const
{
	OP_METHOD (ReturnInsertBetweenLines, 0);
	return this->insert_between_lines;
}

Bool
OPSelGrid::ReturnInsertBetweenColumns () const
{
	OP_METHOD (ReturnInsertBetweenColumns, 0);
	return this->insert_between_columns;
}

Bool
OPSelGrid::ReturnInsertOverCell () const
{
	OP_METHOD (ReturnInsertOverCell, 0);
	return this->insert_over_cell;
}

Count
OPSelGrid::ReturnSelectNum () const
{
	OP_METHOD (ReturnSelectNum, 0);
	return this->select_num;
}


Bool
OPSelGrid::FindActiveHilite (CountOut x, CountOut y, Count nth) const
{
	OP_METHOD (FindActiveHilite, 3);
	OP_ARG (1, CountOut, x);
	OP_ARG (2, CountOut, y);
	OP_ARG (3, Count, nth);
	
	if (nth >= this->select_num) {
		return FALSE;
	}
	
	if (this->search_by_lines) {
		for (y = 0; y < this->v_cell_count; y++) {
			for (x = 0; x < this->h_cell_count; x++) {
				if (this->ReturnActiveHilite (x, y)) {
					if (nth-- == 0) {
						return TRUE;
					}
				}
			}
		}
	} else {
		for (x = 0; x < this->h_cell_count; x++) {
			for (y = 0; y < this->v_cell_count; y++) {
				if (this->ReturnActiveHilite (x, y)) {
					if (nth-- == 0) {
						return TRUE;
					}
				}
			}
		}
	}
	
	return FALSE;
}


void
OPSelGrid::HandleCellClick (OPObjectIn object, Card32 id)
{
	OP_METHOD (HandleCellClick, 2);
	OP_ARG (1, OPObjectIn, object);
	OP_ARG (2, Card32, id);
	
	Count x = id / this->v_cell_count;
	Count y = id - x * this->v_cell_count;
	
	OPViewRef view = object;
	
	if (this->select_on_click) {
		this->SetActiveHilite (x, y, ! this->ReturnActiveHilite (x, y));
	}
}


/*
 *	The button has been pressed in a cell. This will create a
 *	drag view and hand the events over to it...
 */

void
OPSelGrid::HandleDragBegin (OPObjectIn object, Card32 id)
{
	OP_METHOD (HandleDragBegin, 2);
	OP_ARG (1, OPObjectIn, object);
	OP_ARG (2, Card32, id);
	
	if (this->enable_cell_drag) {
		
		Count x = id / this->v_cell_count;
		Count y = id - x * this->v_cell_count;
		
		this->is_dragging   = TRUE;
		this->is_drag_done  = TRUE;
		this->drag_source_h = this->drop_h = x;
		this->drag_source_v = this->drop_v = y;
		
		OPViewRef    cell = object;
		OPBoxViewRef view = OP_NEW (OPBoxView);
		
		const OPRect&     box   = cell->ReturnBox ();
		const OPEventRef& event = OPEvent::CurrentEvent ();
		
		//	Pretend the user released the button (this will be true later on, but
		//	the cell will not get the message, since the release will be sent to
		//	the drag view).
		
		cell->ButtonRelease (event);
		
		view->SetParent (this);
		view->SetOrigin (event->ReturnX () - 5, event->ReturnY () - 5);
		view->SetSize (10, 10);
		view->SetBackgroundColor (OP_COLOR_VIEW_DROPPING);
		view->SetFrameColor (OP_COLOR_BLACK);
		view->EnableTitle (FALSE);
		view->EnableFrame (TRUE);
		
		this->drag_view = OP_NEW (OPDragView);
		
		this->drag_view->SetAttribute ("SourceSelGrid", this);
		this->drag_view->AttachObserver (this, "CancelDrop", "HandleDragCancel");
		this->drag_view->AttachObserver (this, "CommitDrop", "HandleDragCommit");
		this->drag_view->DropOnlyInView (this);
		this->drag_view->SetAutoStop (TRUE);
		this->drag_view->SetAutoOops (FALSE);
		this->drag_view->Setup (view, FALSE);
		
		OPEvent::HandOverToWindow (this->drag_view);
	}
}

void
OPSelGrid::HandleDragCancel ()
{
	this->is_dragging = FALSE;
	this->drag_view->DetachObserver (this, "CancelDrop", "HandleDragCancel");
	this->drag_view->DetachObserver (this, "CommitDrop", "HandleDragCommit");
	this->drag_view->Delete ();
	this->drag_view = OPDragViewNull;
}

void
OPSelGrid::HandleDragCommit ()
{
	OP_ASSERT (this->is_dragging);
	
	OPCoord x = 0;
	OPCoord y = 0;
	OPControlRef cell;
	
	this->drag_view->GetDropOrigin (x, y);
	
	if (this->Detect (x, y, cell)) {
		if (this->FindCell (cell, this->drop_h, this->drop_v)) {
			this->is_drop_done = TRUE;
			this->Report (notify_CellDragged);
		}
	}
	
	this->HandleDragCancel ();
}


/*
 *	Handle the drag and drop protocol in a special way if
 *	we are going to drop a cell in the grid...
 */

Bool
OPSelGrid::DragAndDrop (OPRectIn box, OPObjectIn initiator, OPDragDropProtocol ddp)
{
	OP_METHOD (DragAndDrop, 3);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPObjectIn, initiator);
	OP_ARG (3, OPDragDropProtocol, ddp);
	
	OPObjectRef object;
	
	//	Check to see whether the initiator is a drag view which has been
	//	created by another selection grid. If not, just let the super-class
	//	handle the operation.
	
	if (initiator->GetAttribute ("SourceSelGrid", object) == FALSE) {
		return super::DragAndDrop (box, initiator, ddp);
	}
	
	if ( (ddp == OP_DDP_DRAGnDROP_MOTION)
	  || (ddp == OP_DDP_DRAGnDROP_ENTER) ) {
		
		OPControlRef cell;
		OPCoord x = (box.X1 () + box.X2 ()) / 2;
		OPCoord y = (box.Y1 () + box.Y2 ()) / 2;
		
		if (this->Detect (x, y, cell)) {
			
			Count h = this->drop_h;
			Count v = this->drop_v;
			
			if (this->FindCell (cell, h, v)) {
				
				OPRect rect;
				Bool   draw = FALSE;
				
				cell->GetBox (rect);
				
				OPCoord ox = (rect.X1 () + rect.X2 ()) / 2;
				OPCoord oy = (rect.Y1 () + rect.Y2 ()) / 2;
				OPCoord dx = (rect.DX ()) / 2;
				OPCoord dy = (rect.DY ()) / 2;
				
				this->h_percent = this->insert_between_columns ? (x - ox) / dx : 0;
				this->v_percent = this->insert_between_lines   ? (y - oy) / dy : 0;
				
				if (this->is_drop_valid == FALSE) {
					this->is_drop_valid = TRUE;
					this->window->NotifyInsertOverlay (this);
					draw = TRUE;
				}
				
				this->drop_h = h;
				this->drop_v = v;
				
				//	Find out which kind of overlay will have to be drawn (either a box,
				//	a vertical mark or an horizontal mark).
				
				Real hp = this->h_percent;
				Real vp = this->v_percent;
				
				int overlay = OVERLAY_NONE;
				
				if (hp < 0) { hp = -hp; }
				if (vp < 0) { vp = -vp; }
				
				if ( (hp < 0.3) && (vp < 0.3) && (this->insert_over_cell) ) {
					overlay = OVERLAY_BOX;
				} else if (hp > vp) {
					overlay = this->h_percent < 0 ? OVERLAY_LEFT : OVERLAY_RIGHT;
				} else {
					overlay = this->v_percent < 0 ? OVERLAY_ABOVE : OVERLAY_BELOW;
				}
				
				if (this->draw_overlay != overlay) {
					this->draw_overlay = overlay;
					draw = TRUE;
				}
				
				if (draw) {
					this->window->Display ();
				}
			}
		}
	}
	
	if ( (ddp == OP_DDP_DRAGnDROP_COMMIT)
	  || (ddp == OP_DDP_DRAGnDROP_LEAVE) ) {
		if (this->is_drop_valid) {
			this->is_drop_valid = FALSE;
			this->window->NotifyRemoveOverlay (this);
			this->window->Display ();
		}
	}
	
	//	Handle the substitutions.
	
	if ( (ddp == OP_DDP_DRAGnDROP_SUBST)
	  || (ddp == OP_DDP_CONNECT_SUBST)
	  || (ddp == OP_DDP_UI_EDIT_SUBST) ) {
			
		//	Tell the initiator to use the cell grid rather than the cell as a drop
		//	target...
		
		if (this->parent->DragAndDrop (box, initiator, ddp)) {
			
			static OPMessageRef doSubst = OP_NEW_FROM_ZONE (OPaC_temp_zone, OPMessage);
			static OPAtomRef    method  = "SubstituteTarget";
			
			doSubst->SetMethod (method);
			
			initiator <- doSubst << this;
		}
		
		return FALSE;
	}
	
	return super::DragAndDrop (box, initiator, ddp);
}


/*
 *	Return the position of the cell where a drop has
 *	happened. Return FALSE if no drop has been done.
 */

Bool
OPSelGrid::GetDropPosition (CountOut h, CountOut v, RealOut ph, RealOut pv) const
{
	OP_METHOD (GetDropPosition, 4);
	OP_ARG (1, CountOut, h);
	OP_ARG (2, CountOut, v);
	OP_ARG (3, RealOut, ph);
	OP_ARG (4, RealOut, pv);
	
	if (this->is_drop_done) {
		
		h  = this->drop_h;
		v  = this->drop_v;
		ph = this->h_percent;
		pv = this->v_percent;
		
		return TRUE;
	}
	
	return FALSE;
}

/*
 *	Return the position of the cell where the drag and
 *	drop operation started. Return FALSE if no drag has
 *	been done.
 */

Bool
OPSelGrid::GetDragOrigin (CountOut h, CountOut v) const
{
	OP_METHOD (GetDragOrigin, 2);
	OP_ARG (1, CountOut, h);
	OP_ARG (2, CountOut, v);
	
	if (this->is_drag_done) {
		
		h = this->drag_source_h;
		v = this->drag_source_v;
		
		return TRUE;
	}
	
	return FALSE;
}

void
OPSelGrid::ResetDragDropStatus ()
{
	OP_METHOD (ResetDragDropStatus, 0);
	
	this->is_drop_done = FALSE;
	this->is_drag_done = FALSE;
}


/*
 *	When the cell grid is a drag target, draw the insertion point
 *	afterwards.
 */

void
OPSelGrid::DrawOverlay (OPGrailPortIn port, OPRectIn rect)
{
	OPCoord ox = 0;
	OPCoord oy = 0;
	OPCoord dx = 0;
	OPCoord dy = 0;
	OPRect  box;
	
	this->GetCellPos (this->drop_h, this->drop_v, ox, oy);
	this->GetCellSize (this->drop_h, this->drop_v, dx, dy);
	
	OPAdornerRef adorner = port->ReturnAdorner ();
	
	switch (this->draw_overlay) {
		
		case OVERLAY_BOX:
			box.SetOrigin (ox, oy);
			box.SetSize (dx, dy);
			
			adorner->DrawRect (box, OP_COLOR_VIEW_DROPPING, -1);
			adorner->DrawRect (box, OP_COLOR_VIEW_DROPPING, -2);
			return;
		
		case OVERLAY_LEFT:
			box.SetOrigin (ox-1, oy);
			box.SetSize (2, dy);
			break;
		case OVERLAY_RIGHT:
			box.SetOrigin (ox+dx-1, oy);
			box.SetSize (2, dy);
			break;
		case OVERLAY_ABOVE:
			box.SetOrigin (ox, oy-1);
			box.SetSize (dx, 2);
			break;
		case OVERLAY_BELOW:
			box.SetOrigin (ox, oy+dy-1);
			box.SetSize (dx, 2);
			break;
		
		default:
			return;
	}
	
	adorner->DrawRect (box, OP_COLOR_VIEW_DROPPING);
}

