/*
 *	dlinkviewer.cxx
 *
 *	The data link viewer is a tool which allows to inspect the
 *	links established between source data fields and target data
 *	fields.
 *	
 *	(C) Copyright 1997-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include "buildedit.h"
#include "dlinkviewer.h"

#include <OPaC/boxview.h>
#include <OPaC/button.h>
#include <OPaC/cellgrid.h>
#include <OPaC/database.h>
#include <OPaC/datarep.h>
#include <OPaC/new.h>
#include <OPaC/object.h>
#include <OPaC/pubmacro.h>
#include <OPaC/scrgrid.h>
#include <OPaC/selgrid.h>
#include <OPaC/switch.h>
#include <OPaC/targetmacro.h>
#include <OPaC/tablist.h>
#include <OPaC/tabview.h>
#include <OPaC/textcaption.h>
#include <OPaC/textline.h>
#include <OPaC/textstyle.h>
#include <OPaC/texttag.h>
#include <OPaC/values.h>
#include <OPaC/view.h>
#include <OPaC/window.h>

/****************************************************************/

APP_IMPLEMENT_CLASS (OPDataLinkViewer);

OP_TARGET_ACTOR (OPDataLinkViewer, DisconnectTarget);
OP_TARGET_ACTOR (OPDataLinkViewer, SourceCellActivated);
OP_TARGET_ACTOR (OPDataLinkViewer, TargetCellActivated);
OP_TARGET_ACTOR (OPDataLinkViewer, Update);

/****************************************************************/

OPDataLinkViewer::OPDataLinkViewer ()
{
	OP_METHOD (OPDataLinkViewer, 0);
	this->source_selected = 0;
}

OPDataLinkViewer::~OPDataLinkViewer ()
{
	OP_METHOD (~OPDataLinkViewer, 0);
}

/****************************************************************/

/*
 *	Create a data link viewer for the specified object. This
 *	will handle two scroll lists and a few buttons.
 */

void
OPDataLinkViewer::Initialise (OPObjectIn object, OPStringIn title_source,
							  OPStringIn title_target)
{
	this->object = object;
	
	this->view_source = OP_NEW (OPDBScrollList);
	this->view_target = OP_NEW (OPDBScrollList);
	
	OPCoord x, y;
	OPCoord dx, dy;
	
	this->GetOrigin (x, y);
	this->GetSize (dx, dy);
	
	x  += 5;
	y  += 5;
	dx -= 20;
	dx /= 2;
	dy -= 40;
	
	this->view_source->SetParent (this);
	this->view_source->SetOrigin (x, y);
	this->view_source->SetSize (dx, dy);
	this->view_source->SetupDataBase (title_source);
	this->view_source->AttachObserver (this, "SelectionChanged", "SourceCellActivated");
	this->view_source->ReturnGrid ()->SetSelectAtLeastOne (TRUE);
	this->view_source->Update ();
	
	this->view_target->SetParent (this);
	this->view_target->SetOrigin (x+dx+10, y);
	this->view_target->SetSize (dx, dy);
	this->view_target->SetupDataBase (title_target, FALSE);
	this->view_target->AttachObserver (this, "SelectionChanged", "TargetCellActivated");
	this->view_target->Update ();
	
	OPBuildEdit::MakeDataSourceList (this->view_source, this->object);
	
	this->button_disconnect = OP_NEW (OPButton);
	this->button_disconnect->Initialise (x+2*dx+10-80, y+dy+5, 80, 24, "Disconnect", this);
	this->button_disconnect->AttachObserver (this, "OnClick", "DisconnectTarget");
	
	this->view_source->SetSelectedCell (0);
}


/*
 *	Update the target list according to the selected source
 *	in the left scroll list.
 */

Bool
OPDataLinkViewer::Update ()
{
	OPAtomRef   notifier;
	OPStringRef text;
	OPAtomRef   find;
	
	this->view_target->RemoveAll ();
	
	if ( (this->view_source->GetSelectedText (text))
	  && (this->view_source->GetSelectedName (find))
	  && (text->GetAttribute ("Notifier", notifier)) ) {
		
		Count num   = this->object->CountAttached ();
		Count index = 0;
		
		while (index < num) {
			
			OPAtomRef   event;
			OPAtomRef   actor;
			OPObjectRef observer;
			Card32      id;
			
			this->object->GetAttached (index, event, actor, observer, id);
			
			if ( (event == notifier)
			  && (observer->IsKindOf (OP_RUNTIME (OPDataReport))) ) {
				
				OPDataReportRef reporter = observer;
				
				OPAtomRef   source = reporter->ReturnQueryName ();
				OPAtomRef   name   = reporter->ReturnTargetName ();
				OPObjectRef target = reporter->ReturnTargetObject ();
				
				if (name.IsInvalid ()) {
					name = reporter->ReturnTargetPath ();
				}
				
				if ( (reporter->IsVisible ())
				  && (source == find)) {
					
					OPStringRef text   = "";
					
					text->AppendString (target->ReturnRuntime ()->ReturnClassName ());
					text->AppendString ("::");
					text->AppendString (name);
					
					this->view_target->Insert (text, name, observer, text);
				}
			}
			
			index++;
		}
	}
	
	OPBuildEdit::MakeDataSourceList (this->view_source, this->object);
	
	this->view_source->SetSelectedCell (this->source_selected, FALSE);
	this->view_source->Display ();
	this->view_target->Update ();
	this->view_target->Display ();
	
	return TRUE;
}


/*
 *	A cell has been highlighted in the event list. Update
 *	the actor list accordingly.
 */

Bool
OPDataLinkViewer::SourceCellActivated ()
{
	Count selected = this->source_selected;
	this->view_source->GetSelectedCell (selected);
	
	if (selected == this->source_selected) {
		return TRUE;
	}
	
	this->source_selected = selected;
	this->Update ();
	
	return TRUE;
}


/*
 *	A cell has been highlighted in the actor list. This
 *	might display a description of the selected observer.
 */

Bool
OPDataLinkViewer::TargetCellActivated ()
{
	this->UpdateSelectedInfo ();
	
	//	NIY
	
	return TRUE;
}


/*
 *	The user has clicked on the disconnection button. This
 *	will just remove the highlighted action attached to the
 *	event shown in the left scroll list.
 */

Bool
OPDataLinkViewer::DisconnectTarget ()
{
	OP_METHOD (DisconnectTarget, 0);
	
	this->UpdateSelectedInfo ();
	
	OPDataReportRef reporter;
	
	if ( (this->view_source->CountSelectedItems ())
	  && (this->view_target->CountSelectedItems ())
	  && (this->view_target->GetSelectedData (reporter)) ) {
		
		reporter->Delete ();
		this->Update ();
	}
	
	return TRUE;
}


/*
 *	Update internal information about the selected items
 *	in both scroll lists.
 */

Bool
OPDataLinkViewer::UpdateSelectedInfo ()
{
	OP_METHOD (UpdateSelectedInfo, 0);
	
	this->source_name   = OPAtomNull;
	this->notifier_name = OPAtomNull;
	this->target_name   = OPAtomNull;
	
	if (this->view_source->CountSelectedItems ()) {
		OPStringRef text;
		this->view_source->GetSelectedName (this->source_name);
		this->view_source->GetSelectedText (text);
		text->GetAttribute ("Notifier", this->notifier_name);
	}
	
	if ( (this->view_source->CountSelectedItems ())
	  && (this->view_target->CountSelectedItems ()) ) {
		this->view_target->GetSelectedName (this->target_name);
	}
	
	return TRUE;
}


/*
 *	Return the selected event in the left scroll list.
 */

Bool
OPDataLinkViewer::GetSelectedSource (OPAtomOut source, OPAtomOut notifier)
{
	this->UpdateSelectedInfo ();
	
	source   = this->source_name;
	notifier = this->notifier_name;
	
	return TRUE;
}


/*
 *	Return the selected observer/actor in the right
 *	scroll list.
 */

Bool
OPDataLinkViewer::GetSelectedTarget (OPAtomOut target)
{
	this->UpdateSelectedInfo ();
	
	target = this->target_name;
	
	return TRUE;
}

